# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as exception
from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import excutils
import six

from vmware_nsx._i18n import _
from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import utils
from vmware_nsx.nsxlib import mh as nsxlib
from vmware_nsx.nsxlib.mh import switch
from vmware_nsx.nsxlib.mh import versioning

# @versioning.versioned decorator makes the apparent function body
# totally unrelated to the real function.  This confuses pylint :(
# pylint: disable=assignment-from-no-return

HTTP_GET = "GET"
HTTP_POST = "POST"
HTTP_DELETE = "DELETE"
HTTP_PUT = "PUT"

LROUTER_RESOURCE = "lrouter"
LROUTER_RESOURCE = "lrouter"
LROUTERPORT_RESOURCE = "lport/%s" % LROUTER_RESOURCE
LROUTERRIB_RESOURCE = "rib/%s" % LROUTER_RESOURCE
LROUTERNAT_RESOURCE = "nat/lrouter"
# Constants for NAT rules
MATCH_KEYS = ["destination_ip_addresses", "destination_port_max",
              "destination_port_min", "source_ip_addresses",
              "source_port_max", "source_port_min", "protocol"]

LOG = log.getLogger(__name__)


def _prepare_lrouter_body(name, neutron_router_id, tenant_id,
                          router_type, distributed=None, **kwargs):
    body = {
        "display_name": utils.check_and_truncate(name),
        "tags": utils.get_tags(os_tid=tenant_id,
                               q_router_id=neutron_router_id),
        "routing_config": {
            "type": router_type
        },
        "type": "LogicalRouterConfig",
        "replication_mode": cfg.CONF.NSX.replication_mode,
    }
    # add the distributed key only if not None (ie: True or False)
    if distributed is not None:
        body['distributed'] = distributed
    if kwargs:
        body["routing_config"].update(kwargs)
    return body


def _create_implicit_routing_lrouter(cluster, neutron_router_id, tenant_id,
                                     display_name, nexthop, distributed=None):
    implicit_routing_config = {
        "default_route_next_hop": {
            "gateway_ip_address": nexthop,
            "type": "RouterNextHop"
        },
    }
    lrouter_obj = _prepare_lrouter_body(
        display_name, neutron_router_id, tenant_id,
        "SingleDefaultRouteImplicitRoutingConfig",
        distributed=distributed,
        **implicit_routing_config)
    return nsxlib.do_request(HTTP_POST,
                             nsxlib._build_uri_path(LROUTER_RESOURCE),
                             jsonutils.dumps(lrouter_obj), cluster=cluster)


def create_implicit_routing_lrouter(cluster, neutron_router_id, tenant_id,
                                    display_name, nexthop):
    """Create a NSX logical router on the specified cluster.

        :param cluster: The target NSX cluster
        :param tenant_id: Identifier of the Openstack tenant for which
        the logical router is being created
        :param display_name: Descriptive name of this logical router
        :param nexthop: External gateway IP address for the logical router
        :raise NsxApiException: if there is a problem while communicating
        with the NSX controller
    """
    return _create_implicit_routing_lrouter(
        cluster, neutron_router_id, tenant_id, display_name, nexthop)


def create_implicit_routing_lrouter_with_distribution(
    cluster, neutron_router_id, tenant_id, display_name,
    nexthop, distributed=None):
    """Create a NSX logical router on the specified cluster.

    This function also allows for creating distributed lrouters
    :param cluster: The target NSX cluster
    :param tenant_id: Identifier of the Openstack tenant for which
    the logical router is being created
    :param display_name: Descriptive name of this logical router
    :param nexthop: External gateway IP address for the logical router
    :param distributed: True for distributed logical routers
    :raise NsxApiException: if there is a problem while communicating
    with the NSX controller
    """
    return _create_implicit_routing_lrouter(
        cluster, neutron_router_id, tenant_id,
        display_name, nexthop, distributed)


def create_explicit_routing_lrouter(cluster, neutron_router_id, tenant_id,
                                    display_name, nexthop, distributed=None):
    lrouter_obj = _prepare_lrouter_body(
        display_name, neutron_router_id, tenant_id,
        "RoutingTableRoutingConfig", distributed=distributed)
    router = nsxlib.do_request(HTTP_POST,
                               nsxlib._build_uri_path(LROUTER_RESOURCE),
                               jsonutils.dumps(lrouter_obj), cluster=cluster)
    default_gw = {'prefix': '0.0.0.0/0', 'next_hop_ip': nexthop}
    create_explicit_route_lrouter(cluster, router['uuid'], default_gw)
    return router


def delete_lrouter(cluster, lrouter_id):
    nsxlib.do_request(HTTP_DELETE,
                      nsxlib._build_uri_path(LROUTER_RESOURCE,
                                             resource_id=lrouter_id),
                      cluster=cluster)


def get_lrouter(cluster, lrouter_id):
    return nsxlib.do_request(HTTP_GET,
                             nsxlib._build_uri_path(
                                 LROUTER_RESOURCE,
                                 resource_id=lrouter_id,
                                 relations='LogicalRouterStatus'),
                             cluster=cluster)


def query_lrouters(cluster, fields=None, filters=None):
    return nsxlib.get_all_query_pages(
        nsxlib._build_uri_path(LROUTER_RESOURCE,
                               fields=fields,
                               relations='LogicalRouterStatus',
                               filters=filters),
        cluster)


def get_lrouters(cluster, tenant_id, fields=None, filters=None):
    # FIXME(salv-orlando): Fields parameter is ignored in this routine
    actual_filters = {}
    if filters:
        actual_filters.update(filters)
    if tenant_id:
        actual_filters['tag'] = tenant_id
        actual_filters['tag_scope'] = 'os_tid'
    lrouter_fields = "uuid,display_name,fabric_status,tags"
    return query_lrouters(cluster, lrouter_fields, actual_filters)


def update_implicit_routing_lrouter(cluster, r_id, display_name, nexthop):
    lrouter_obj = get_lrouter(cluster, r_id)
    if not display_name and not nexthop:
        # Nothing to update
        return lrouter_obj
    # It seems that this is faster than the doing an if on display_name
    lrouter_obj["display_name"] = (utils.check_and_truncate(display_name) or
                                   lrouter_obj["display_name"])
    if nexthop:
        nh_element = lrouter_obj["routing_config"].get(
            "default_route_next_hop")
        if nh_element:
            nh_element["gateway_ip_address"] = nexthop
    return nsxlib.do_request(HTTP_PUT,
                             nsxlib._build_uri_path(LROUTER_RESOURCE,
                                                    resource_id=r_id),
                             jsonutils.dumps(lrouter_obj),
                             cluster=cluster)


def get_explicit_routes_lrouter(cluster, router_id, protocol_type='static'):
    static_filter = {'protocol': protocol_type}
    existing_routes = nsxlib.do_request(
        HTTP_GET,
        nsxlib._build_uri_path(LROUTERRIB_RESOURCE,
                               filters=static_filter,
                               fields="*",
                               parent_resource_id=router_id),
        cluster=cluster)['results']
    return existing_routes


def delete_explicit_route_lrouter(cluster, router_id, route_id):
    nsxlib.do_request(HTTP_DELETE,
                      nsxlib._build_uri_path(LROUTERRIB_RESOURCE,
                                             resource_id=route_id,
                                             parent_resource_id=router_id),
                      cluster=cluster)


def create_explicit_route_lrouter(cluster, router_id, route):
    next_hop_ip = route.get("nexthop") or route.get("next_hop_ip")
    prefix = route.get("destination") or route.get("prefix")
    uuid = nsxlib.do_request(
        HTTP_POST,
        nsxlib._build_uri_path(LROUTERRIB_RESOURCE,
                               parent_resource_id=router_id),
        jsonutils.dumps({
            "action": "accept",
            "next_hop_ip": next_hop_ip,
            "prefix": prefix,
            "protocol": "static"
        }),
        cluster=cluster)['uuid']
    return uuid


def update_explicit_routes_lrouter(cluster, router_id, routes):
    # Update in bulk: delete them all, and add the ones specified
    # but keep track of what is been modified to allow roll-backs
    # in case of failures
    nsx_routes = get_explicit_routes_lrouter(cluster, router_id)
    try:
        deleted_routes = []
        added_routes = []
        # omit the default route (0.0.0.0/0) from the processing;
        # this must be handled through the nexthop for the router
        for route in nsx_routes:
            prefix = route.get("destination") or route.get("prefix")
            if prefix != '0.0.0.0/0':
                delete_explicit_route_lrouter(cluster,
                                              router_id,
                                              route['uuid'])
                deleted_routes.append(route)
        for route in routes:
            prefix = route.get("destination") or route.get("prefix")
            if prefix != '0.0.0.0/0':
                uuid = create_explicit_route_lrouter(cluster,
                                                     router_id, route)
                added_routes.append(uuid)
    except api_exc.NsxApiException:
        LOG.exception('Cannot update NSX routes %(routes)s for '
                      'router %(router_id)s',
                      {'routes': routes, 'router_id': router_id})
        # Roll back to keep NSX in consistent state
        with excutils.save_and_reraise_exception():
            if nsx_routes:
                if deleted_routes:
                    for route in deleted_routes:
                        create_explicit_route_lrouter(cluster,
                                                      router_id, route)
                if added_routes:
                    for route_id in added_routes:
                        delete_explicit_route_lrouter(cluster,
                                                      router_id, route_id)
    return nsx_routes


def get_default_route_explicit_routing_lrouter_v33(cluster, router_id):
    static_filter = {"protocol": "static",
                     "prefix": "0.0.0.0/0"}
    default_route = nsxlib.do_request(
        HTTP_GET,
        nsxlib._build_uri_path(LROUTERRIB_RESOURCE,
                               filters=static_filter,
                               fields="*",
                               parent_resource_id=router_id),
        cluster=cluster)["results"][0]
    return default_route


def get_default_route_explicit_routing_lrouter_v32(cluster, router_id):
    # Scan all routes because 3.2 does not support query by prefix
    all_routes = get_explicit_routes_lrouter(cluster, router_id)
    for route in all_routes:
        if route['prefix'] == '0.0.0.0/0':
            return route


def update_default_gw_explicit_routing_lrouter(cluster, router_id, next_hop):
    default_route = get_default_route_explicit_routing_lrouter(cluster,
                                                               router_id)
    if next_hop != default_route["next_hop_ip"]:
        new_default_route = {"action": "accept",
                             "next_hop_ip": next_hop,
                             "prefix": "0.0.0.0/0",
                             "protocol": "static"}
        nsxlib.do_request(HTTP_PUT,
                          nsxlib._build_uri_path(
                              LROUTERRIB_RESOURCE,
                              resource_id=default_route['uuid'],
                              parent_resource_id=router_id),
                          jsonutils.dumps(new_default_route),
                          cluster=cluster)


def update_explicit_routing_lrouter(cluster, router_id,
                                    display_name, next_hop, routes=None):
    update_implicit_routing_lrouter(cluster, router_id, display_name, next_hop)
    if next_hop:
        update_default_gw_explicit_routing_lrouter(cluster,
                                                   router_id, next_hop)
    if routes is not None:
        return update_explicit_routes_lrouter(cluster, router_id, routes)


def query_lrouter_lports(cluster, lr_uuid, fields="*",
                         filters=None, relations=None):
    uri = nsxlib._build_uri_path(LROUTERPORT_RESOURCE,
                                 parent_resource_id=lr_uuid,
                                 fields=fields, filters=filters,
                                 relations=relations)
    return nsxlib.do_request(HTTP_GET, uri, cluster=cluster)['results']


def create_router_lport(cluster, lrouter_uuid, tenant_id, neutron_port_id,
                        display_name, admin_status_enabled, ip_addresses,
                        mac_address=None):
    """Creates a logical port on the assigned logical router."""
    lport_obj = dict(
        admin_status_enabled=admin_status_enabled,
        display_name=display_name,
        tags=utils.get_tags(os_tid=tenant_id, q_port_id=neutron_port_id),
        ip_addresses=ip_addresses,
        type="LogicalRouterPortConfig"
    )
    # Only add the mac_address to lport_obj if present. This is because
    # when creating the fake_ext_gw there is no mac_address present.
    if mac_address:
        lport_obj['mac_address'] = mac_address
    path = nsxlib._build_uri_path(LROUTERPORT_RESOURCE,
                                  parent_resource_id=lrouter_uuid)
    result = nsxlib.do_request(HTTP_POST, path, jsonutils.dumps(lport_obj),
                               cluster=cluster)

    LOG.debug("Created logical port %(lport_uuid)s on "
              "logical router %(lrouter_uuid)s",
              {'lport_uuid': result['uuid'],
               'lrouter_uuid': lrouter_uuid})
    return result


def update_router_lport(cluster, lrouter_uuid, lrouter_port_uuid,
                        tenant_id, neutron_port_id, display_name,
                        admin_status_enabled, ip_addresses):
    """Updates a logical port on the assigned logical router."""
    lport_obj = dict(
        admin_status_enabled=admin_status_enabled,
        display_name=display_name,
        tags=utils.get_tags(os_tid=tenant_id, q_port_id=neutron_port_id),
        ip_addresses=ip_addresses,
        type="LogicalRouterPortConfig"
    )
    # Do not pass null items to NSX
    for key in lport_obj.keys():
        if lport_obj[key] is None:
            del lport_obj[key]
    path = nsxlib._build_uri_path(LROUTERPORT_RESOURCE,
                                  lrouter_port_uuid,
                                  parent_resource_id=lrouter_uuid)
    result = nsxlib.do_request(HTTP_PUT, path,
                               jsonutils.dumps(lport_obj),
                               cluster=cluster)
    LOG.debug("Updated logical port %(lport_uuid)s on "
              "logical router %(lrouter_uuid)s",
              {'lport_uuid': lrouter_port_uuid, 'lrouter_uuid': lrouter_uuid})
    return result


def delete_router_lport(cluster, lrouter_uuid, lport_uuid):
    """Creates a logical port on the assigned logical router."""
    path = nsxlib._build_uri_path(LROUTERPORT_RESOURCE, lport_uuid,
                                  lrouter_uuid)
    nsxlib.do_request(HTTP_DELETE, path, cluster=cluster)
    LOG.debug("Delete logical router port %(lport_uuid)s on "
              "logical router %(lrouter_uuid)s",
              {'lport_uuid': lport_uuid,
               'lrouter_uuid': lrouter_uuid})


def delete_peer_router_lport(cluster, lr_uuid, ls_uuid, lp_uuid):
    nsx_port = switch.get_port(cluster, ls_uuid, lp_uuid,
                               relations="LogicalPortAttachment")
    relations = nsx_port.get('_relations')
    if relations:
        att_data = relations.get('LogicalPortAttachment')
        if att_data:
            lrp_uuid = att_data.get('peer_port_uuid')
            if lrp_uuid:
                delete_router_lport(cluster, lr_uuid, lrp_uuid)


def find_router_gw_port(context, cluster, router_id):
    """Retrieves the external gateway port for a NSX logical router."""

    # Find the uuid of nsx ext gw logical router port
    # TODO(salvatore-orlando): Consider storing it in Neutron DB
    results = query_lrouter_lports(
        cluster, router_id,
        relations="LogicalPortAttachment")
    for lport in results:
        if '_relations' in lport:
            attachment = lport['_relations'].get('LogicalPortAttachment')
            if attachment and attachment.get('type') == 'L3GatewayAttachment':
                return lport


def plug_router_port_attachment(cluster, router_id, port_id,
                                attachment_uuid, nsx_attachment_type,
                                attachment_vlan=None):
    """Attach a router port to the given attachment.

    Current attachment types:
       - PatchAttachment [-> logical switch port uuid]
       - L3GatewayAttachment [-> L3GatewayService uuid]
    For the latter attachment type a VLAN ID can be specified as well.
    """
    uri = nsxlib._build_uri_path(LROUTERPORT_RESOURCE, port_id, router_id,
                                 is_attachment=True)
    attach_obj = {}
    attach_obj["type"] = nsx_attachment_type
    if nsx_attachment_type == "PatchAttachment":
        attach_obj["peer_port_uuid"] = attachment_uuid
    elif nsx_attachment_type == "L3GatewayAttachment":
        attach_obj["l3_gateway_service_uuid"] = attachment_uuid
        if attachment_vlan:
            attach_obj['vlan_id'] = attachment_vlan
    else:
        raise nsx_exc.InvalidAttachmentType(
            attachment_type=nsx_attachment_type)
    return nsxlib.do_request(
        HTTP_PUT, uri, jsonutils.dumps(attach_obj), cluster=cluster)


def _create_nat_match_obj(**kwargs):
    nat_match_obj = {'ethertype': 'IPv4'}
    delta = set(kwargs.keys()) - set(MATCH_KEYS)
    if delta:
        raise Exception(_("Invalid keys for NAT match: %s"), delta)
    nat_match_obj.update(kwargs)
    return nat_match_obj


def _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj):
    LOG.debug("Creating NAT rule: %s", nat_rule_obj)
    uri = nsxlib._build_uri_path(LROUTERNAT_RESOURCE,
                                 parent_resource_id=router_id)
    return nsxlib.do_request(HTTP_POST, uri, jsonutils.dumps(nat_rule_obj),
                             cluster=cluster)


def _build_snat_rule_obj(min_src_ip, max_src_ip, nat_match_obj):
    return {"to_source_ip_address_min": min_src_ip,
            "to_source_ip_address_max": max_src_ip,
            "type": "SourceNatRule",
            "match": nat_match_obj}


def create_lrouter_nosnat_rule_v2(cluster, _router_id, _match_criteria=None):
    LOG.info("No SNAT rules cannot be applied as they are not available "
             "in this version of the NSX platform")


def create_lrouter_nodnat_rule_v2(cluster, _router_id, _match_criteria=None):
    LOG.info("No DNAT rules cannot be applied as they are not available "
             "in this version of the NSX platform")


def create_lrouter_snat_rule_v2(cluster, router_id,
                                min_src_ip, max_src_ip, match_criteria=None):

    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = _build_snat_rule_obj(min_src_ip, max_src_ip, nat_match_obj)
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def create_lrouter_dnat_rule_v2(cluster, router_id, dst_ip,
                                to_dst_port=None, match_criteria=None):

    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = {
        "to_destination_ip_address_min": dst_ip,
        "to_destination_ip_address_max": dst_ip,
        "type": "DestinationNatRule",
        "match": nat_match_obj
    }
    if to_dst_port:
        nat_rule_obj['to_destination_port'] = to_dst_port
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def create_lrouter_nosnat_rule_v3(cluster, router_id, order=None,
                                  match_criteria=None):
    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = {
        "type": "NoSourceNatRule",
        "match": nat_match_obj
    }
    if order:
        nat_rule_obj['order'] = order
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def create_lrouter_nodnat_rule_v3(cluster, router_id, order=None,
                                  match_criteria=None):
    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = {
        "type": "NoDestinationNatRule",
        "match": nat_match_obj
    }
    if order:
        nat_rule_obj['order'] = order
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def create_lrouter_snat_rule_v3(cluster, router_id, min_src_ip, max_src_ip,
                                order=None, match_criteria=None):
    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = _build_snat_rule_obj(min_src_ip, max_src_ip, nat_match_obj)
    if order:
        nat_rule_obj['order'] = order
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def create_lrouter_dnat_rule_v3(cluster, router_id, dst_ip, to_dst_port=None,
                                order=None, match_criteria=None):

    nat_match_obj = _create_nat_match_obj(**match_criteria)
    nat_rule_obj = {
        "to_destination_ip_address": dst_ip,
        "type": "DestinationNatRule",
        "match": nat_match_obj
    }
    if to_dst_port:
        nat_rule_obj['to_destination_port'] = to_dst_port
    if order:
        nat_rule_obj['order'] = order
    return _create_lrouter_nat_rule(cluster, router_id, nat_rule_obj)


def delete_nat_rules_by_match(cluster, router_id, rule_type,
                              max_num_expected,
                              min_num_expected=0,
                              raise_on_len_mismatch=True,
                              **kwargs):
    # remove nat rules
    nat_rules = query_nat_rules(cluster, router_id)
    to_delete_ids = []
    for r in nat_rules:
        if (r['type'] != rule_type):
            continue

        for key, value in six.iteritems(kwargs):
            if not (key in r['match'] and r['match'][key] == value):
                break
        else:
            to_delete_ids.append(r['uuid'])
    num_rules_to_delete = len(to_delete_ids)
    if (num_rules_to_delete < min_num_expected or
        num_rules_to_delete > max_num_expected):
        if raise_on_len_mismatch:
            raise nsx_exc.NatRuleMismatch(actual_rules=num_rules_to_delete,
                                          min_rules=min_num_expected,
                                          max_rules=max_num_expected)
        else:
            LOG.warning("Found %(actual_rule_num)d matching NAT rules, "
                        "which is not in the expected range "
                        "(%(min_exp_rule_num)d,%(max_exp_rule_num)d)",
                        {'actual_rule_num': num_rules_to_delete,
                         'min_exp_rule_num': min_num_expected,
                         'max_exp_rule_num': max_num_expected})

    for rule_id in to_delete_ids:
        delete_router_nat_rule(cluster, router_id, rule_id)
    # Return number of deleted rules - useful at least for
    # testing purposes
    return num_rules_to_delete


def delete_router_nat_rule(cluster, router_id, rule_id):
    uri = nsxlib._build_uri_path(LROUTERNAT_RESOURCE, rule_id, router_id)
    nsxlib.do_request(HTTP_DELETE, uri, cluster=cluster)


def query_nat_rules(cluster, router_id, fields="*", filters=None):
    uri = nsxlib._build_uri_path(LROUTERNAT_RESOURCE,
                                 parent_resource_id=router_id,
                                 fields=fields, filters=filters)
    return nsxlib.get_all_query_pages(uri, cluster)


# NOTE(salvatore-orlando): The following FIXME applies in general to
# each operation on list attributes.
# FIXME(salvatore-orlando): need a lock around the list of IPs on an iface
def update_lrouter_port_ips(cluster, lrouter_id, lport_id,
                            ips_to_add, ips_to_remove):
    uri = nsxlib._build_uri_path(LROUTERPORT_RESOURCE, lport_id, lrouter_id)
    try:
        port = nsxlib.do_request(HTTP_GET, uri, cluster=cluster)
        # TODO(salvatore-orlando): Enforce ips_to_add intersection with
        # ips_to_remove is empty
        ip_address_set = set(port['ip_addresses'])
        ip_address_set = ip_address_set - set(ips_to_remove)
        ip_address_set = ip_address_set | set(ips_to_add)
        # Set is not JSON serializable - convert to list
        port['ip_addresses'] = list(ip_address_set)
        nsxlib.do_request(HTTP_PUT, uri, jsonutils.dumps(port),
                          cluster=cluster)
    except exception.NotFound:
        # FIXME(salv-orlando):avoid raising different exception
        data = {'lport_id': lport_id, 'lrouter_id': lrouter_id}
        msg = (_("Router Port %(lport_id)s not found on router "
                 "%(lrouter_id)s") % data)
        LOG.exception(msg)
        raise nsx_exc.NsxPluginException(err_msg=msg)
    except api_exc.NsxApiException as e:
        msg = _("An exception occurred while updating IP addresses on a "
                "router logical port:%s") % e
        LOG.exception(msg)
        raise nsx_exc.NsxPluginException(err_msg=msg)


ROUTER_FUNC_DICT = {
    'create_lrouter': {
        2: {versioning.DEFAULT_VERSION: create_implicit_routing_lrouter, },
        3: {versioning.DEFAULT_VERSION: create_implicit_routing_lrouter,
            1: create_implicit_routing_lrouter_with_distribution,
            2: create_explicit_routing_lrouter, }, },
    'update_lrouter': {
        2: {versioning.DEFAULT_VERSION: update_implicit_routing_lrouter, },
        3: {versioning.DEFAULT_VERSION: update_implicit_routing_lrouter,
            2: update_explicit_routing_lrouter, }, },
    'create_lrouter_dnat_rule': {
        2: {versioning.DEFAULT_VERSION: create_lrouter_dnat_rule_v2, },
        3: {versioning.DEFAULT_VERSION: create_lrouter_dnat_rule_v3, }, },
    'create_lrouter_snat_rule': {
        2: {versioning.DEFAULT_VERSION: create_lrouter_snat_rule_v2, },
        3: {versioning.DEFAULT_VERSION: create_lrouter_snat_rule_v3, }, },
    'create_lrouter_nosnat_rule': {
        2: {versioning.DEFAULT_VERSION: create_lrouter_nosnat_rule_v2, },
        3: {versioning.DEFAULT_VERSION: create_lrouter_nosnat_rule_v3, }, },
    'create_lrouter_nodnat_rule': {
        2: {versioning.DEFAULT_VERSION: create_lrouter_nodnat_rule_v2, },
        3: {versioning.DEFAULT_VERSION: create_lrouter_nodnat_rule_v3, }, },
    'get_default_route_explicit_routing_lrouter': {
        3: {versioning.DEFAULT_VERSION:
            get_default_route_explicit_routing_lrouter_v32,
            2: get_default_route_explicit_routing_lrouter_v32, }, },
}


@versioning.versioned(ROUTER_FUNC_DICT)
def create_lrouter(cluster, *args, **kwargs):
    if kwargs.get('distributed', None):
        v = cluster.api_client.get_version()
        if (v.major, v.minor) < (3, 1):
            raise nsx_exc.InvalidVersion(version=v)
        return v


@versioning.versioned(ROUTER_FUNC_DICT)
def get_default_route_explicit_routing_lrouter(cluster, *args, **kwargs):
    pass


@versioning.versioned(ROUTER_FUNC_DICT)
def update_lrouter(cluster, *args, **kwargs):
    if kwargs.get('routes', None):
        v = cluster.api_client.get_version()
        if (v.major, v.minor) < (3, 2):
            raise nsx_exc.InvalidVersion(version=v)
        return v


@versioning.versioned(ROUTER_FUNC_DICT)
def create_lrouter_dnat_rule(cluster, *args, **kwargs):
    pass


@versioning.versioned(ROUTER_FUNC_DICT)
def create_lrouter_snat_rule(cluster, *args, **kwargs):
    pass


@versioning.versioned(ROUTER_FUNC_DICT)
def create_lrouter_nosnat_rule(cluster, *args, **kwargs):
    pass


@versioning.versioned(ROUTER_FUNC_DICT)
def create_lrouter_nodnat_rule(cluster, *args, **kwargs):
    pass



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import mock
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit.network.v2 import test_port
from openstackclient.tests.unit import utils as tests_utils

from vmware_nsx.osc.v2 import port


supported_extensions = ('vnic-index',
                        'provider-security-group',
                        'mac-learning')


class TestCreatePort(test_port.TestCreatePort):

    def setUp(self):
        super(TestCreatePort, self).setUp()
        # Get the command object to test
        self.cmd = port.NsxCreatePort(self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_create_with_arg_and_val(self, arg_name, arg_val, is_valid=True):
        self.network.create_port.reset_mock()
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            self._port.name,
            '--network', self._port.network_id,
            conv_name, str(arg_val)
        ]
        verifylist = [
            ('name', self._port.name),
            ('network', self._port.network_id,),
            (arg_name, arg_val),
            ('enable', True),
        ]
        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = (self.cmd.take_action(parsed_args))
        self.network.create_port.assert_called_once_with(**{
            'admin_state_up': True,
            'network_id': self._port.network_id,
            'name': self._port.name,
            arg_name: arg_val,
        })

        ref_columns, ref_data = self._get_common_cols_data(self._port)
        self.assertEqual(ref_columns, columns)
        self.assertEqual(ref_data, data)

    def _test_create_with_vnix_index(self, val, is_valid=True):
        self._test_create_with_arg_and_val('vnic_index', val, is_valid)

    def test_create_with_vnic_index(self):
        self._test_create_with_vnix_index(1)

    def test_create_with_illegal_vnic_index(self):
        self._test_create_with_vnix_index('illegal', is_valid=False)

    def test_create_with_provider_security_group(self):
        # create a port with 1 provider security group
        secgroup = network_fakes.FakeSecurityGroup.create_one_security_group()
        self.network.find_security_group = mock.Mock(return_value=secgroup)
        arglist = [
            '--network', self._port.network_id,
            '--provider-security-group', secgroup.id,
            'test-port',
        ]
        verifylist = [
            ('network', self._port.network_id,),
            ('enable', True),
            ('provider_security_groups', [secgroup.id]),
            ('name', 'test-port'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_port.assert_called_once_with(**{
            'admin_state_up': True,
            'network_id': self._port.network_id,
            'provider_security_groups': [secgroup.id],
            'name': 'test-port',
        })

        ref_columns, ref_data = self._get_common_cols_data(self._port)
        self.assertEqual(ref_columns, columns)
        self.assertEqual(ref_data, data)

    def test_create_with_provider_security_groups(self):
        # create a port with few provider security groups
        sg_1 = network_fakes.FakeSecurityGroup.create_one_security_group()
        sg_2 = network_fakes.FakeSecurityGroup.create_one_security_group()
        self.network.find_security_group = mock.Mock(side_effect=[sg_1, sg_2])
        arglist = [
            '--network', self._port.network_id,
            '--provider-security-group', sg_1.id,
            '--provider-security-group', sg_2.id,
            'test-port',
        ]
        verifylist = [
            ('network', self._port.network_id,),
            ('enable', True),
            ('provider_security_groups', [sg_1.id, sg_2.id]),
            ('name', 'test-port'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_port.assert_called_once_with(**{
            'admin_state_up': True,
            'network_id': self._port.network_id,
            'provider_security_groups': [sg_1.id, sg_2.id],
            'name': 'test-port',
        })

        ref_columns, ref_data = self._get_common_cols_data(self._port)
        self.assertEqual(ref_columns, columns)
        self.assertEqual(ref_data, data)

    def test_create_with_provider_security_group_by_name(self):
        # create a port with 1 provider security group
        secgroup = network_fakes.FakeSecurityGroup.create_one_security_group()
        self.network.find_security_group = mock.Mock(return_value=secgroup)
        arglist = [
            '--network', self._port.network_id,
            '--provider-security-group', secgroup.name,
            'test-port',
        ]
        verifylist = [
            ('network', self._port.network_id,),
            ('enable', True),
            ('provider_security_groups', [secgroup.name]),
            ('name', 'test-port'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_port.assert_called_once_with(**{
            'admin_state_up': True,
            'network_id': self._port.network_id,
            'provider_security_groups': [secgroup.id],
            'name': 'test-port',
        })

        ref_columns, ref_data = self._get_common_cols_data(self._port)
        self.assertEqual(ref_columns, columns)
        self.assertEqual(ref_data, data)

    def _test_create_with_flag_arg(
        self, arg_name, validate_name, validate_val):
        self.network.create_port.reset_mock()
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            self._port.name,
            '--network', self._port.network_id,
            conv_name
        ]
        verifylist = [
            ('name', self._port.name),
            ('network', self._port.network_id,),
            (arg_name, True),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = (self.cmd.take_action(parsed_args))
        self.network.create_port.assert_called_once_with(**{
            'admin_state_up': True,
            'network_id': self._port.network_id,
            'name': self._port.name,
            validate_name: validate_val,
        })

        ref_columns, ref_data = self._get_common_cols_data(self._port)
        self.assertEqual(ref_columns, columns)
        self.assertEqual(ref_data, data)

    def test_create_with_mac_learning(self):
        self._test_create_with_flag_arg(
            'enable_mac_learning', 'mac_learning_enabled', True)

    def test_create_with_no_mac_learning(self):
        self._test_create_with_flag_arg(
            'disable_mac_learning', 'mac_learning_enabled', False)


class TestSetPort(test_port.TestSetPort):

    def setUp(self):
        super(TestSetPort, self).setUp()
        # Get the command object to test
        self.cmd = port.NsxSetPort(self.app, self.namespace)
        # mock the relevant extensions
        get_ext = mock.patch('vmware_nsx.osc.v2.utils.get_extensions').start()
        get_ext.return_value = supported_extensions

    def _test_set_with_arg_and_val(self, arg_name, arg_val, is_valid=True):
        self.network.update_port.reset_mock()
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            self._port.name,
            conv_name, str(arg_val)
        ]
        verifylist = [
            ('port', self._port.name),
            (arg_name, arg_val)
        ]
        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {arg_name: arg_val}
        self.network.update_port.assert_called_once_with(
            self._port, **attrs)
        self.assertIsNone(result)

    def _test_set_Vnic_index(self, val, is_valid=True):
        self._test_set_with_arg_and_val('vnic_index', val, is_valid)

    def test_set_vnic_index(self):
        self._test_set_Vnic_index(1)

    def test_set_illegal_vnic_index(self):
        # check illegal index
        self._test_set_Vnic_index('illegal', is_valid=False)

    def test_set_provider_security_group(self):
        # It is not allowed to change the provider security groups
        sg = network_fakes.FakeSecurityGroup.create_one_security_group()
        self.network.find_security_group = mock.Mock(return_value=sg)
        arglist = [
            '--provider-security-group', sg.id,
            self._port.name,
        ]
        verifylist = [
            ('provider_security_groups', [sg.id]),
            ('port', self._port.name),
        ]

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)

    def _test_set_with_flag_arg(self, arg_name, validate_name,
                                validate_val, is_valid=True):
        self.network.update_port.reset_mock()
        # add '--' to the arg name and change '_' to '-'
        conv_name = '--' + re.sub('_', '-', arg_name)
        arglist = [
            self._port.name,
            conv_name
        ]
        verifylist = [
            ('port', self._port.name),
            (arg_name, True)
        ]
        if not is_valid:
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
            return

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {validate_name: validate_val}
        self.network.update_port.assert_called_once_with(
            self._port, **attrs)
        self.assertIsNone(result)

    def test_set_with_mac_learning(self):
        self._test_set_with_flag_arg(
            'enable_mac_learning', 'mac_learning_enabled', True)

    def test_set_with_no_mac_learning(self):
        self._test_set_with_flag_arg(
            'disable_mac_learning', 'mac_learning_enabled', False)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import getopt
import logging
import re
import sys
import xml.etree.ElementTree as et

from keystoneauth1 import identity
from keystoneauth1 import session
import libvirt
from neutronclient.v2_0 import client
import nova.conf

CONF = nova.conf.CONF
logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)


def usage():
    print ("python nsx_instance_if_migrate.py --username=<username> "
           "--password=<password> --project=<project> "
           "--auth-url=<keystone auth URL> "
           "[--project-domain-id=<project domain>] "
           "[--user-domain-id=<user domain>] "
           "[--machine-type=<migrated machine type] "
           "[--logfile=<log file>] "
           "[--nsx-bridge=<NSX managed vSwitch>]\n\n"
           "Convert libvirt interface definitions on a KVM host, to NSX "
           "managed vSwitch definitions\n\n"
           "  username: Admin user's username\n"
           "  password: Admin user's password\n"
           "  keystone auth URL: URL to keystone's authentication service\n"
           "  project domain: Keystone project domain\n"
           "  user domain: Keystone user domain\n"
           "  migrated machine type: Overwrites libvirt's machine type\n"
           "  log file: Output log of the command execution\n"
           "  NSX managed vSwitch: vSwitch on host, managed by NSX\n\n")

    sys.exit()


def get_opts():
    opts = {}
    o = []
    p = re.compile('^-+')
    try:
        o, a = getopt.getopt(sys.argv[1:], 'h', ['help',
                                                 'username=',
                                                 'password=',
                                                 'project=',
                                                 'project-domain-id=',
                                                 'user-domain-id=',
                                                 'auth-url=',
                                                 'machine-type=',
                                                 'logfile=',
                                                 'nsx-bridge='])
    except getopt.GetoptError as err:
        LOG.error(err)
        usage()
    for opt, val in o:
        if opt in ('h', 'help'):
            usage()
        else:
            opts[p.sub('', opt)] = val

    for mandatory_key in ['username', 'password', 'project', 'auth-url']:
        if opts.get(mandatory_key) is None:
            LOG.error("%s must be specified!", mandatory_key)
            usage()

    return opts


def xmltag_text_get(obj, tag_name):
    tag_obj = obj.find(tag_name)
    if tag_obj is not None:
        return tag_obj.text


def xmltag_attr_get(obj, tag, attr):
    tag_obj = obj.find(tag)
    if tag_obj is not None:
        return tag_obj.get(attr)


def xmltag_set(elem, tag, **kwargs):
    sub_elem = elem.find(tag)
    if sub_elem is None:
        sub_elem = et.SubElement(elem, tag)
    for attr in kwargs.keys():
        sub_elem.set(attr, kwargs.get(attr))
    return sub_elem


def iface_migrate(neutron, instance_name, iface, nsx_switch):
    iface.set('type', 'bridge')
    xmltag_set(iface, 'source', bridge=nsx_switch)
    virt_port = xmltag_set(iface, 'virtualport', type='openvswitch')
    instance_mac = xmltag_attr_get(iface, 'mac', 'address')
    if instance_mac is None:
        LOG.error("Couldn't find MAC address for instance %s", instance_name)
        return

    ports = neutron.list_ports(fields=['id'], mac_address=instance_mac)
    if len(ports['ports']) != 1:
        LOG.error('For instance %(vm)s, invalid ports received from neutron: '
                  '%(ports)s', {'vm': instance_name, 'ports': ports})
        return

    neutron_port_id = ports['ports'][0]['id']
    xmltag_set(virt_port, 'parameters', interfaceid=neutron_port_id)
    xmltag_set(iface, 'driver', name='qemu')

    tap_dev = xmltag_attr_get(iface, 'target', 'dev')
    if tap_dev is None:
        LOG.error("For instance %(vm)s, couldn't find tap device for "
                  "interface", instance_name)

    # remove script tag if found
    script_tag = iface.find('script')
    if script_tag is not None:
        iface.remove(script_tag)


def is_valid_os_data(libvirt_conn, os_type, os_arch, os_machine):
    caps_xml = libvirt_conn.getCapabilities()
    caps_root = et.fromstring(caps_xml)
    for guest_tag in caps_root.findall('guest'):
        if (xmltag_text_get(guest_tag, 'os_type') == os_type
           and xmltag_attr_get(guest_tag, 'arch', 'name') == os_arch):
            for machine_tag in guest_tag.find('arch').findall('machine'):
                if machine_tag.text == os_machine:
                    return True
    return False


def instance_migrate(libvirt_conn, neutron, instance, machine_type,
                     nsx_switch):
    xml = instance.XMLDesc()
    root = et.fromstring(xml)

    instance_name = xmltag_text_get(root, 'name')
    if instance_name is None:
        LOG.error("Couldn't find instance name in XML")
        return

    instance_uuid = xmltag_text_get(root, 'uuid')
    if instance_uuid is None:
        LOG.error("Couldn't find UUID for instance %s", instance_name)
        return

    # Validate that os is supported by hypervisor
    os_tag = root.find('os')
    if os_tag is None:
        LOG.error("Couldn't find OS tag for instance %s", instance_name)
        return
    type_tag = os_tag.find('type')
    if not is_valid_os_data(libvirt_conn, type_tag.text, type_tag.get('arch'),
                            type_tag.get('machine')):
        LOG.error("Instance %s OS data is invalid or not supported by "
                  "hypervisor", instance_name)
        return

    if machine_type is not None:
        type_tag.set('machine', machine_type)

    devs = root.find('devices')
    ifaces = devs.findall('interface')

    if not ifaces:
        LOG.error('No interfaces to migrate for instance %s', instance_name)

    for iface in ifaces:
        iface_migrate(neutron, instance_name, iface, nsx_switch)

    instance.undefine()
    libvirt_conn.defineXML(et.tostring(root))
    LOG.info('Migrated instance %(vm)s (%(uuid)s) successfully!',
             {'vm': instance_name, 'uuid': instance_uuid})


def main():
    opts = get_opts()
    if opts.get('logfile'):
        f_handler = logging.FileHandler(opts.get('logfile'))
        f_formatter = logging.Formatter(
            '%(asctime)s %(levelname)s %(message)s')
        f_handler.setFormatter(f_formatter)
        LOG.addHandler(f_handler)

    conn = libvirt.open('qemu:///system')
    if conn is None:
        LOG.error('Failed to connect to libvirt')
        exit(1)

    auth = identity.Password(username=opts['username'],
                             password=opts['password'],
                             project_name=opts['project'],
                             project_domain_id=opts.get('project-domain-id',
                                                        'default'),
                             user_domain_id=opts.get('user-domain-id',
                                                     'default'),
                             auth_url=opts['auth-url'])

    if auth is None:
        LOG.error('Failed to authenticate with keystone')
        exit(1)

    sess = session.Session(auth=auth)
    if sess is None:
        LOG.error('Failed to create keystone session')
        exit(1)

    neutron = client.Client(session=sess)
    if neutron is None:
        LOG.error('Failed to create neutron session')
        exit(1)

    instances = conn.listAllDomains()
    if not instances:
        LOG.error('No instances to migrate')

    for instance in instances:
        try:
            instance_migrate(conn, neutron, instance, opts.get('machine-type'),
                             opts.get('nsx-bridge', CONF.neutron.ovs_bridge))
        except Exception as e:
            LOG.error('Failed to migrate instance with exception %s', e)


if __name__ == "__main__":
    main()



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api import extensions

LOGGING = 'logging'

RESOURCE_ATTRIBUTE_MAP = {
    'security_groups': {
        LOGGING: {
            'allow_post': True,
            'allow_put': True,
            'convert_to': converters.convert_to_boolean,
            'default': False,
            'enforce_policy': True,
            'is_visible': True}
    }
}


class Securitygrouplogging(extensions.ExtensionDescriptor):
    """Security group logging extension."""

    @classmethod
    def get_name(cls):
        return "Security group logging"

    @classmethod
    def get_alias(cls):
        return "security-group-logging"

    @classmethod
    def get_description(cls):
        return "Security group logging extension."

    @classmethod
    def get_namespace(cls):
        # todo
        return "https://docs.openstack.org/ext/security_group_logging/api/v2.0"

    @classmethod
    def get_updated(cls):
        return "2015-04-13T10:00:00-00:00"

    def get_required_extensions(self):
        return ["security-group"]

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        return []

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}



# Copyright 2012 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from oslo_log import log as logging
from six.moves import http_client as httplib

from vmware_nsx.api_client import base
from vmware_nsx.api_client import eventlet_client
from vmware_nsx.api_client import eventlet_request
from vmware_nsx.api_client import exception
from vmware_nsx.api_client import version

LOG = logging.getLogger(__name__)


class NsxApiClient(eventlet_client.EventletApiClient):
    """The Nsx API Client."""

    def __init__(self, api_providers, user, password,
                 concurrent_connections=base.DEFAULT_CONCURRENT_CONNECTIONS,
                 gen_timeout=base.GENERATION_ID_TIMEOUT,
                 use_https=True,
                 connect_timeout=base.DEFAULT_CONNECT_TIMEOUT,
                 http_timeout=75, retries=2, redirects=2):
        '''Constructor. Adds the following:

        :param http_timeout: how long to wait before aborting an
            unresponsive controller (and allow for retries to another
            controller in the cluster)
        :param retries: the number of concurrent connections.
        :param redirects: the number of concurrent connections.
        '''
        super(NsxApiClient, self).__init__(
            api_providers, user, password,
            concurrent_connections=concurrent_connections,
            gen_timeout=gen_timeout, use_https=use_https,
            connect_timeout=connect_timeout)

        self._request_timeout = http_timeout * retries
        self._http_timeout = http_timeout
        self._retries = retries
        self._redirects = redirects
        self._version = None

    # NOTE(salvatore-orlando): This method is not used anymore. Login is now
    # performed automatically inside the request eventlet if necessary.
    def login(self, user=None, password=None):
        '''Login to NSX controller.

        Assumes same password is used for all controllers.

        :param user: controller user (usually admin). Provided for
                backwards compatibility. In the  normal mode of operation
                this should be None.
        :param password: controller password. Provided for backwards
                compatibility. In the normal mode of operation this should
                be None.
        '''
        if user:
            self._user = user
        if password:
            self._password = password

        return self._login()

    def request(self, method, url, body="", content_type="application/json"):
        '''Issues request to controller.'''

        g = eventlet_request.GenericRequestEventlet(
            self, method, url, body, content_type, auto_login=True,
            http_timeout=self._http_timeout,
            retries=self._retries, redirects=self._redirects)
        g.start()
        response = g.join()
        LOG.debug('Request returns "%s"', response)

        # response is a modified HTTPResponse object or None.
        # response.read() will not work on response as the underlying library
        # request_eventlet.ApiRequestEventlet has already called this
        # method in order to extract the body and headers for processing.
        # ApiRequestEventlet derived classes call .read() and
        # .getheaders() on the HTTPResponse objects and store the results in
        # the response object's .body and .headers data members for future
        # access.

        if response is None:
            # Timeout.
            LOG.error('Request timed out: %(method)s to %(url)s',
                      {'method': method, 'url': url})
            raise exception.RequestTimeout()

        status = response.status
        if status == httplib.UNAUTHORIZED:
            raise exception.UnAuthorizedRequest()

        # Fail-fast: Check for exception conditions and raise the
        # appropriate exceptions for known error codes.
        if status in exception.ERROR_MAPPINGS:
            LOG.error("Received error code: %s", status)
            LOG.error("Server Error Message: %s", response.body)
            exception.ERROR_MAPPINGS[status](response)

        # Continue processing for non-error condition.
        if (status != httplib.OK and status != httplib.CREATED
                and status != httplib.NO_CONTENT):
            LOG.error("%(method)s to %(url)s, unexpected response code: "
                      "%(status)d (content = '%(body)s')",
                      {'method': method, 'url': url,
                       'status': response.status, 'body': response.body})
            return None

        if not self._version:
            self._version = version.find_version(response.headers)
        return response.body

    def get_version(self):
        if not self._version:
            # Determine the controller version by querying the
            # cluster nodes. Currently, the version will be the
            # one of the server that responds.
            self.request('GET', '/ws.v1/control-cluster/node')
            if not self._version:
                LOG.error('Unable to determine NSX version. '
                          'Plugin might not work as expected.')
        return self._version



# Copyright 2017 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib.db import constants as db_constants
from neutron_lib.services.qos import base
from neutron_lib.services.qos import constants as qos_consts
from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap

LOG = logging.getLogger(__name__)
DRIVER = None
SUPPORTED_RULES = {
    qos_consts.RULE_TYPE_BANDWIDTH_LIMIT: {
        qos_consts.MAX_KBPS: {
            'type:range': [0, db_constants.DB_INTEGER_MAX_VALUE]},
        qos_consts.MAX_BURST: {
            'type:range': [0, db_constants.DB_INTEGER_MAX_VALUE]},
        qos_consts.DIRECTION: {
            'type:values': [constants.EGRESS_DIRECTION,
                            constants.INGRESS_DIRECTION]}
    },
    qos_consts.RULE_TYPE_DSCP_MARKING: {
        qos_consts.DSCP_MARK: {'type:values': constants.VALID_DSCP_MARKS}
    }
}


class NSXvQosDriver(base.DriverBase):

    @staticmethod
    def create(core_plugin):
        return NSXvQosDriver(
            core_plugin,
            name='NSXvQosDriver',
            vif_types=None,
            vnic_types=None,
            supported_rules=SUPPORTED_RULES,
            requires_rpc_notifications=False)

    def __init__(self, core_plugin, **kwargs):
        super(NSXvQosDriver, self).__init__(**kwargs)
        self.core_plugin = core_plugin
        if self.core_plugin.is_tvd_plugin():
            # get the plugin that match this driver
            self.core_plugin = self.core_plugin.get_plugin_by_type(
                projectpluginmap.NsxPlugins.NSX_V)
        self.requires_rpc_notifications = False

    def is_vif_type_compatible(self, vif_type):
        return True

    def is_vnic_compatible(self, vnic_type):
        return True

    def create_policy(self, context, policy):
        pass

    def update_policy(self, context, policy):
        # get all the bound networks of this policy
        networks = policy.get_bound_networks()
        for net_id in networks:
            # update the new bw limitations for this network
            self.core_plugin._update_qos_on_backend_network(
                context, net_id, policy.id)

    def delete_policy(self, context, policy):
        pass


def register(core_plugin):
    """Register the NSX-V QoS driver."""
    global DRIVER
    if not DRIVER:
        DRIVER = NSXvQosDriver.create(core_plugin)
    LOG.debug('NSXvQosDriver QoS driver registered')



# Copyright 2012 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from neutron_lib.api.definitions import allowedaddresspairs as addr_apidef
from neutron_lib.api.definitions import port_security as psec
from neutron_lib.exceptions import allowedaddresspairs as addr_exc
from neutron_lib.exceptions import port_security as psec_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.api import extensions as neutron_extensions
from neutron.db import _resource_extend as resource_extend
from neutron.db import _utils as db_utils
from neutron.db import agentschedulers_db
from neutron.db import allowedaddresspairs_db as addr_pair_db
from neutron.db import api as db_api
from neutron.db import dns_db
from neutron.db import external_net_db
from neutron.db import l3_db
from neutron.db.models import securitygroup as securitygroup_model  # noqa
from neutron.db import models_v2
from neutron.db import portbindings_db
from neutron.db import portsecurity_db
from neutron.db import securitygroups_db
from neutron.db import vlantransparent_db as vlan_ext_db
from neutron.extensions import providernet
from neutron.extensions import securitygroup as ext_sg
from neutron.plugins.common import utils
from neutron.quota import resource_registry
from neutron_lib.api.definitions import multiprovidernet as mpnet_apidef
from neutron_lib.api.definitions import portbindings as pbin
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib.api.definitions import vlantransparent as vlan_apidef
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import exceptions as n_exc

import vmware_nsx
from vmware_nsx._i18n import _
from vmware_nsx.common import config  # noqa
from vmware_nsx.common import managers as nsx_managers
from vmware_nsx.common import nsx_constants
from vmware_nsx.common import utils as c_utils
from vmware_nsx.db import db as nsx_db
from vmware_nsx.db import nsxv_db
from vmware_nsx.dhcp_meta import modes as dhcpmeta_modes
from vmware_nsx.dvs import dvs
from vmware_nsx.dvs import dvs_utils
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.common import plugin as nsx_plugin_common
from vmware_nsx.plugins.nsx import utils as tvd_utils

LOG = logging.getLogger(__name__)


@resource_extend.has_resource_extenders
class NsxDvsV2(addr_pair_db.AllowedAddressPairsMixin,
               agentschedulers_db.DhcpAgentSchedulerDbMixin,
               nsx_plugin_common.NsxPluginBase,
               dhcpmeta_modes.DhcpMetadataAccess,
               external_net_db.External_net_db_mixin,
               l3_db.L3_NAT_dbonly_mixin,
               portbindings_db.PortBindingMixin,
               portsecurity_db.PortSecurityDbMixin,
               securitygroups_db.SecurityGroupDbMixin,
               dns_db.DNSDbMixin,
               vlan_ext_db.Vlantransparent_db_mixin):

    supported_extension_aliases = ["allowed-address-pairs",
                                   "binding",
                                   "external-net",
                                   "multi-provider",
                                   "port-security",
                                   "provider",
                                   "quotas",
                                   "router",
                                   "security-group",
                                   "vlan-transparent"]

    __native_bulk_support = True
    __native_pagination_support = True
    __native_sorting_support = True

    @resource_registry.tracked_resources(
        network=models_v2.Network,
        port=models_v2.Port,
        subnet=models_v2.Subnet,
        subnetpool=models_v2.SubnetPool,
        security_group=securitygroup_model.SecurityGroup,
        security_group_rule=securitygroup_model.SecurityGroupRule)
    def __init__(self):
        self._is_sub_plugin = tvd_utils.is_tvd_core_plugin()
        dvs_utils.dvs_register_exceptions()
        super(NsxDvsV2, self).__init__()
        if self._is_sub_plugin:
            extension_drivers = cfg.CONF.nsx_tvd.dvs_extension_drivers
        else:
            extension_drivers = cfg.CONF.nsx_extension_drivers
        self._extension_manager = nsx_managers.ExtensionManager(
             extension_drivers=extension_drivers)
        LOG.debug('Driver support: DVS: %s' % dvs_utils.dvs_is_enabled())
        self._extension_manager.initialize()
        self.supported_extension_aliases.extend(
            self._extension_manager.extension_aliases())
        neutron_extensions.append_api_extensions_path(
            [vmware_nsx.NSX_EXT_PATH])
        self.cfg_group = 'dvs'  # group name for dvs section in nsx.ini
        self._dvs = dvs.SingleDvsManager()
        self.setup_dhcpmeta_access()

    @staticmethod
    def plugin_type():
        return projectpluginmap.NsxPlugins.DVS

    @staticmethod
    def is_tvd_plugin():
        return False

    @staticmethod
    def _extend_port_dict_binding(result, portdb):
        result[pbin.VIF_TYPE] = nsx_constants.VIF_TYPE_DVS
        port_attr = portdb.get('nsx_port_attributes')
        if port_attr:
            result[pbin.VNIC_TYPE] = port_attr.vnic_type
        else:
            result[pbin.VNIC_TYPE] = pbin.VNIC_NORMAL
        result[pbin.VIF_DETAILS] = {
            # TODO(rkukura): Replace with new VIF security details
            # security-groups extension supported by this plugin
            pbin.CAP_PORT_FILTER: True}

    def _extend_get_network_dict_provider(self, context, network,
                                          multiprovider=None, bindings=None):
        if not bindings:
            bindings = nsx_db.get_network_bindings(context.session,
                                                   network['id'])
        if not multiprovider:
            multiprovider = nsx_db.is_multiprovider_network(context.session,
                                                            network['id'])
        # With NSX plugin 'normal' overlay networks will have no binding
        # TODO(salvatore-orlando) make sure users can specify a distinct
        # phy_uuid as 'provider network' for STT net type
        if bindings:
            if not multiprovider:
                # network came in through provider networks api
                network[pnet.NETWORK_TYPE] = bindings[0].binding_type
                network[pnet.PHYSICAL_NETWORK] = bindings[0].phy_uuid
                network[pnet.SEGMENTATION_ID] = bindings[0].vlan_id
            else:
                # network come in though multiprovider networks api
                network[mpnet_apidef.SEGMENTS] = [
                    {pnet.NETWORK_TYPE: binding.binding_type,
                     pnet.PHYSICAL_NETWORK: binding.phy_uuid,
                     pnet.SEGMENTATION_ID: binding.vlan_id}
                    for binding in bindings]

    def _dvs_get_id(self, net_data):
        if net_data['name'] == '':
            return net_data['id']
        else:
            # Maximum name length is 80 characters. 'id' length is 36
            # maximum prefix for name is 43
            return '%s-%s' % (net_data['name'][:43], net_data['id'])

    def _dvs_create_network(self, context, network):
        net_data = network['network']
        if net_data['admin_state_up'] is False:
            LOG.warning("Network with admin_state_up=False are not yet "
                        "supported by this plugin. Ignoring setting for "
                        "network %s", net_data.get('name', '<unknown>'))
        net_data['id'] = str(uuid.uuid4())
        vlan_tag = 0
        if net_data.get(pnet.NETWORK_TYPE) == c_utils.NetworkTypes.VLAN:
            vlan_tag = net_data.get(pnet.SEGMENTATION_ID, 0)

        trunk_mode = False
        # vlan transparent can be an object if not set.
        if net_data.get(vlan_apidef.VLANTRANSPARENT) is True:
            trunk_mode = True

        net_id = None
        if net_data.get(pnet.NETWORK_TYPE) == c_utils.NetworkTypes.PORTGROUP:
            net_id = net_data.get(pnet.PHYSICAL_NETWORK)
            pg_info = self._dvs.get_port_group_info(net_id)
            if pg_info.get('name') != net_data.get('name'):
                err_msg = (_("Portgroup name %(dvpg)s must match network "
                            "name %(network)s") % {'dvpg': pg_info.get('name'),
                            'network': net_data.get('name')})
                raise n_exc.InvalidInput(error_message=err_msg)
            dvpg_moref = self._dvs.net_id_to_moref(net_id)
            dvs_id = dvpg_moref.value
        else:
            dvs_id = self._dvs_get_id(net_data)
            try:
                self._dvs.add_port_group(dvs_id, vlan_tag,
                                         trunk_mode=trunk_mode)
            except dvs_utils.DvsOperationBulkFault:
                LOG.warning('One or more hosts may not be configured')

        try:
            with db_api.context_manager.writer.using(context):
                new_net = super(NsxDvsV2, self).create_network(context,
                                                               network)
                self._extension_manager.process_create_network(
                    context, net_data, new_net)
                # Process port security extension
                self._process_network_port_security_create(
                    context, net_data, new_net)

                # Process vlan transparent extension
                net_db = self._get_network(context, new_net['id'])
                net_db['vlan_transparent'] = trunk_mode
                net_data['vlan_transparent'] = trunk_mode
                resource_extend.apply_funcs('networks', net_data, net_db)

                nsx_db.add_network_binding(
                    context.session, new_net['id'],
                    net_data.get(pnet.NETWORK_TYPE),
                    net_id or 'dvs',
                    vlan_tag)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Failed to create network')
                if (net_data.get(pnet.NETWORK_TYPE) !=
                        c_utils.NetworkTypes.PORTGROUP):
                    self._dvs.delete_port_group(dvs_id)

        new_net[pnet.NETWORK_TYPE] = net_data.get(pnet.NETWORK_TYPE)
        new_net[pnet.PHYSICAL_NETWORK] = net_id or 'dvs'
        new_net[pnet.SEGMENTATION_ID] = vlan_tag

        # this extra lookup is necessary to get the
        # latest db model for the extension functions
        net_model = self._get_network(context, net_data['id'])
        resource_extend.apply_funcs('networks', new_net, net_model)

        self.handle_network_dhcp_access(context, new_net,
                                        action='create_network')
        return new_net

    def _validate_network(self, context, net_data):
        network_type = net_data.get(pnet.NETWORK_TYPE)
        segmentation_id = net_data.get(pnet.SEGMENTATION_ID)
        segmentation_id_set = validators.is_attr_set(segmentation_id)
        if not context.is_admin:
            err_msg = _("Only an admin can create a DVS provider "
                        "network")
            raise n_exc.InvalidInput(error_message=err_msg)
        err_msg = None
        if (network_type == c_utils.NetworkTypes.FLAT or
            network_type == c_utils.NetworkTypes.PORTGROUP):
            if segmentation_id_set:
                err_msg = (_("Segmentation ID cannot be specified with "
                            "%s network type"), network_type)
        elif network_type == c_utils.NetworkTypes.VLAN:
            if not segmentation_id_set:
                err_msg = _("Segmentation ID must be specified with "
                            "vlan network type")
            if (segmentation_id_set and
                not utils.is_valid_vlan_tag(segmentation_id)):
                err_msg = (_("%(segmentation_id)s out of range "
                             "(%(min_id)s through %(max_id)s)") %
                           {'segmentation_id': segmentation_id,
                            'min_id': constants.MIN_VLAN_TAG,
                            'max_id': constants.MAX_VLAN_TAG})
        else:
            err_msg = (_("%(net_type_param)s %(net_type_value)s not "
                         "supported") %
                       {'net_type_param': pnet.NETWORK_TYPE,
                        'net_type_value': network_type})
        if err_msg:
            raise n_exc.InvalidInput(error_message=err_msg)

    def create_network(self, context, network):
        self._validate_network(context, network['network'])
        return self._dvs_create_network(context, network)

    def _dvs_delete_network(self, context, id):
        network = self._get_network(context, id)
        dvs_id = self._dvs_get_id(network)
        bindings = nsx_db.get_network_bindings(context.session, id)
        with db_api.context_manager.writer.using(context):
            nsx_db.delete_network_bindings(context.session, id)
            super(NsxDvsV2, self).delete_network(context, id)
        try:
            if (not bindings or
                bindings[0].binding_type != c_utils.NetworkTypes.PORTGROUP):
                self._dvs.delete_port_group(dvs_id)
        except Exception:
            LOG.exception('Unable to delete DVS port group %s', id)
        self.handle_network_dhcp_access(context, id, action='delete_network')

    def delete_network(self, context, id):
        self._dvs_delete_network(context, id)

    def _dvs_get_network(self, context, id, fields=None):
        with db_api.context_manager.reader.using(context):
            # goto to the plugin DB and fetch the network
            network = self._get_network(context, id)
            # Don't do field selection here otherwise we won't be able
            # to add provider networks fields
            net_result = self._make_network_dict(network,
                                                 context=context)
            self._extend_get_network_dict_provider(context, net_result)
        return db_utils.resource_fields(net_result, fields)

    def get_network(self, context, id, fields=None):
        return self._dvs_get_network(context, id, fields=None)

    def get_networks(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None,
                     page_reverse=False):
        filters = filters or {}
        with db_api.context_manager.reader.using(context):
            networks = (
                super(NsxDvsV2, self).get_networks(
                    context, filters, fields, sorts,
                    limit, marker, page_reverse))
            for net in networks:
                self._extend_get_network_dict_provider(context, net)
        return (networks if not fields else
                [db_utils.resource_fields(network,
                                          fields) for network in networks])

    def update_network(self, context, id, network):
        net_attrs = network['network']
        providernet._raise_if_updates_provider_attributes(net_attrs)

        with db_api.context_manager.writer.using(context):
            net_res = super(NsxDvsV2, self).update_network(context, id,
                                                           network)
            self._extension_manager.process_update_network(context, net_attrs,
                                                           net_res)
            # Process port security extension
            self._process_network_port_security_update(
                context, net_attrs, net_res)
            self._extend_get_network_dict_provider(context, net_res)

        return net_res

    def _process_vnic_type(self, context, port_data, port_id):
        vnic_type = port_data.get(pbin.VNIC_TYPE)
        if validators.is_attr_set(vnic_type):
            if (vnic_type != pbin.VNIC_NORMAL and
                vnic_type != pbin.VNIC_DIRECT and
                vnic_type != pbin.VNIC_DIRECT_PHYSICAL):
                err_msg = _("Only direct, direct-physical and normal VNIC "
                            "types supported")
                raise n_exc.InvalidInput(error_message=err_msg)
            nsxv_db.update_nsxv_port_ext_attributes(
                session=context.session,
                port_id=port_id,
                vnic_type=vnic_type)

    def create_port(self, context, port):
        # If PORTSECURITY is not the default value ATTR_NOT_SPECIFIED
        # then we pass the port to the policy engine. The reason why we don't
        # pass the value to the policy engine when the port is
        # ATTR_NOT_SPECIFIED is for the case where a port is created on a
        # shared network that is not owned by the tenant.
        port_data = port['port']

        with db_api.context_manager.writer.using(context):
            # First we allocate port in neutron database
            neutron_db = super(NsxDvsV2, self).create_port(context, port)
            self._extension_manager.process_create_port(
                context, port_data, neutron_db)
            port_security = self._get_network_security_binding(
                context, neutron_db['network_id'])
            port_data[psec.PORTSECURITY] = port_security
            self._process_port_port_security_create(
                context, port_data, neutron_db)
            # Update fields obtained from neutron db (eg: MAC address)
            port["port"].update(neutron_db)
            has_ip = self._ip_on_port(neutron_db)

            # security group extension checks
            if has_ip:
                self._ensure_default_security_group_on_port(context, port)
            elif validators.is_attr_set(port_data.get(ext_sg.SECURITYGROUPS)):
                raise psec_exc.PortSecurityAndIPRequiredForSecurityGroups()
            port_data[ext_sg.SECURITYGROUPS] = (
                self._get_security_groups_on_port(context, port))
            self._process_port_create_security_group(
                context, port_data, port_data[ext_sg.SECURITYGROUPS])
            self._process_portbindings_create_and_update(context,
                                                         port['port'],
                                                         port_data)

            # allowed address pair checks
            if validators.is_attr_set(port_data.get(
                    addr_apidef.ADDRESS_PAIRS)):
                if not port_security:
                    raise addr_exc.AddressPairAndPortSecurityRequired()
                else:
                    self._process_create_allowed_address_pairs(
                        context, neutron_db,
                        port_data[addr_apidef.ADDRESS_PAIRS])
            else:
                # remove ATTR_NOT_SPECIFIED
                port_data[addr_apidef.ADDRESS_PAIRS] = []

            self._process_portbindings_create_and_update(context,
                                                         port['port'],
                                                         port_data)
            self._process_vnic_type(context, port_data, neutron_db['id'])

            LOG.debug("create_port completed on NSX for tenant "
                      "%(tenant_id)s: (%(id)s)", port_data)

        # DB Operation is complete, perform DVS operation
        port_data = port['port']

        # this extra lookup is necessary to get the
        # latest db model for the extension functions
        port_model = self._get_port(context, port_data['id'])
        resource_extend.apply_funcs('ports', port_data, port_model)
        self._extend_port_dict_binding(port_data, port_model)

        self.handle_port_dhcp_access(context, port_data, action='create_port')
        return port_data

    def update_port(self, context, id, port):
        delete_addr_pairs = self._check_update_deletes_allowed_address_pairs(
            port)
        has_addr_pairs = self._check_update_has_allowed_address_pairs(port)

        with db_api.context_manager.writer.using(context):
            ret_port = super(NsxDvsV2, self).update_port(
                context, id, port)
            # Save current mac learning state to check whether it's
            # being updated or not
            # copy values over - except fixed_ips as
            # they've already been processed
            port['port'].pop('fixed_ips', None)
            ret_port.update(port['port'])

            # populate port_security setting
            if psec.PORTSECURITY not in port['port']:
                ret_port[psec.PORTSECURITY] = self._get_port_security_binding(
                    context, id)
            # validate port security and allowed address pairs
            if not ret_port[psec.PORTSECURITY]:
                #  has address pairs in request
                if has_addr_pairs:
                    raise addr_exc.AddressPairAndPortSecurityRequired()
                elif not delete_addr_pairs:
                    # check if address pairs are in db
                    ret_port[addr_apidef.ADDRESS_PAIRS] = (
                        self.get_allowed_address_pairs(context, id))
                    if ret_port[addr_apidef.ADDRESS_PAIRS]:
                        raise addr_exc.AddressPairAndPortSecurityRequired()

            if delete_addr_pairs or has_addr_pairs:
                # delete address pairs and read them in
                self._delete_allowed_address_pairs(context, id)
                self._process_create_allowed_address_pairs(
                    context, ret_port, ret_port[addr_apidef.ADDRESS_PAIRS])

            if psec.PORTSECURITY in port['port']:
                self._process_port_port_security_update(
                    context, port['port'], ret_port)
            self._process_vnic_type(context, port['port'], id)
            LOG.debug("Updating port: %s", port)
            self._extension_manager.process_update_port(
                context, port['port'], ret_port)
            self._process_portbindings_create_and_update(context,
                                                         port['port'],
                                                         ret_port)
        return ret_port

    def delete_port(self, context, id, l3_port_check=True,
                    nw_gw_port_check=True):
        """Deletes a port on a specified Virtual Network.

        If the port contains a remote interface attachment, the remote
        interface is first un-plugged and then the port is deleted.

        :returns: None
        :raises: exception.PortInUse
        :raises: exception.PortNotFound
        :raises: exception.NetworkNotFound
        """
        neutron_db_port = self.get_port(context, id)

        with db_api.context_manager.writer.using(context):
            # metadata_dhcp_host_route
            self.handle_port_metadata_access(
                context, neutron_db_port, is_delete=True)
            super(NsxDvsV2, self).delete_port(context, id)
        self.handle_port_dhcp_access(
            context, neutron_db_port, action='delete_port')

    def get_ports(self, context, filters=None, fields=None,
                  sorts=None, limit=None, marker=None,
                  page_reverse=False):
        filters = filters or {}
        with db_api.context_manager.reader.using(context):
            ports = (
                super(NsxDvsV2, self).get_ports(
                      context, filters, fields, sorts,
                      limit, marker, page_reverse))
            # Add port extensions
            for port in ports:
                if 'id' in port:
                    port_model = self._get_port(context, port['id'])
                    resource_extend.apply_funcs('ports', port, port_model)
                    self._extend_port_dict_binding(port, port_model)
        return (ports if not fields else
                [db_utils.resource_fields(port, fields) for port in ports])

    def get_port(self, context, id, fields=None):
        port = super(NsxDvsV2, self).get_port(context, id, fields=None)
        if 'id' in port:
            port_model = self._get_port(context, port['id'])
            resource_extend.apply_funcs('ports', port, port_model)
            self._extend_port_dict_binding(port, port_model)
        else:
            port[pbin.VIF_TYPE] = nsx_constants.VIF_TYPE_DVS
        return db_utils.resource_fields(port, fields)

    def create_router(self, context, router):
        # DVS backend cannot support logical router
        msg = (_("Unable to create router %s with DVS") %
               router['router']['name'])
        raise n_exc.BadRequest(resource="router", msg=msg)

    def get_network_availability_zones(self, net_db):
        """Api to comply with the NSX-TVD plugin"""
        return []



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api import extensions

# Attribute Map
VNIC_INDEX = 'vnic_index'


EXTENDED_ATTRIBUTES_2_0 = {
    'ports': {
        VNIC_INDEX:
        {'allow_post': True,
         'allow_put': True,
         'is_visible': True,
         'default': None,
         'convert_to': converters.convert_to_int_if_not_none}}}


class Vnicindex(extensions.ExtensionDescriptor):
    @classmethod
    def get_name(cls):
        return "VNIC Index"

    @classmethod
    def get_alias(cls):
        return "vnic-index"

    @classmethod
    def get_description(cls):
        return ("Enable a port to be associated with a VNIC index")

    @classmethod
    def get_updated(cls):
        return "2014-09-15T12:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lib.exceptions import firewall_v1 as exceptions

from vmware_nsx.services.fwaas.nsx_v3 import edge_fwaas_driver_base as \
    base_driver

LOG = logging.getLogger(__name__)
FWAAS_DRIVER_NAME = 'Fwaas V1 NSX-V3 driver'
NSX_FW_TAG = 'os-neutron-fw-id'


class EdgeFwaasV3DriverV1(base_driver.CommonEdgeFwaasV3Driver):
    """NSX-V3 driver for Firewall As A Service - V1."""

    def __init__(self):
        exception_cls = exceptions.FirewallInternalDriverError
        super(EdgeFwaasV3DriverV1, self).__init__(exception_cls,
                                                  FWAAS_DRIVER_NAME)

    @log_helpers.log_method_call
    def create_firewall(self, agent_mode, apply_list, firewall):
        """Create the Firewall with a given policy. """
        self._update_backend_routers(apply_list, firewall['id'])

    @log_helpers.log_method_call
    def update_firewall(self, agent_mode, apply_list, firewall):
        """Remove previous policy and apply the new policy."""
        self._update_backend_routers(apply_list, firewall['id'])

    @log_helpers.log_method_call
    def delete_firewall(self, agent_mode, apply_list, firewall):
        """Delete firewall.

        Removes rules created by this instance from the backend firewall
        And add the default allow rule.
        """
        self._update_backend_routers(apply_list, firewall['id'])

    @log_helpers.log_method_call
    def apply_default_policy(self, agent_mode, apply_list, firewall):
        """Apply the default policy (deny all).

        The backend firewall always has this policy (=deny all) as default,
        so we only need to delete the current rules.
        """
        self._update_backend_routers(apply_list, firewall['id'])

    def _update_backend_routers(self, apply_list, fw_id):
        """"Update each router on the backend using the core plugin code"""
        self.validate_backend_version()
        context = n_context.get_admin_context()
        for router_info in apply_list:
            # Skip unsupported routers
            if not self.should_apply_firewall_to_router(router_info.router):
                continue

            self.core_plugin.update_router_firewall(
                context, router_info.router_id)

    def update_nsx_router_tags(self, nsx_router_id, fw_id=None):
        """Update the backend router with tags marking the attached fw id"""
        # Get the current tags
        nsx_router = self.nsx_router.get(nsx_router_id)
        if 'tags' not in nsx_router:
            nsx_router['tags'] = []
        tags = nsx_router['tags']

        # Look for the firewall tag and update/remove it
        update_tags = False
        found_tag = False
        for tag in tags:
            if tag.get('scope') == NSX_FW_TAG:
                found_tag = True
                if not fw_id:
                    tags.remove(tag)
                    update_tags = True
                    break
                if fw_id != tag.get('tag'):
                    tag['tag'] = fw_id
                    update_tags = True
                    break
        # Add the tag if not found
        if fw_id and not found_tag:
            tags.append({'scope': NSX_FW_TAG,
                         'tag': fw_id})
            update_tags = True

        # update tags on the backend router
        if update_tags:
            self.nsx_router.update(nsx_router_id, tags=tags)

    def get_router_translated_rules(self, router_id, firewall):
        """Return the list of translated rules

        The default drop all will be added later
        """
        # Return the firewall rules only if the fw is up
        if firewall['admin_state_up']:
            # TODO(asarfaty): get this value from the firewall extensions
            logged = False
            return self._translate_rules(firewall['firewall_rule_list'],
                                         logged=logged)

        return []



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import exceptions as nc_exc
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas.nsx_v3 import healthmonitor_mgr as hm_mgr
from vmware_nsx.services.lbaas.nsx_v3 import l7policy_mgr
from vmware_nsx.services.lbaas.nsx_v3 import l7rule_mgr
from vmware_nsx.services.lbaas.nsx_v3 import listener_mgr
from vmware_nsx.services.lbaas.nsx_v3 import loadbalancer_mgr as lb_mgr
from vmware_nsx.services.lbaas.nsx_v3 import member_mgr
from vmware_nsx.services.lbaas.nsx_v3 import pool_mgr

LOG = logging.getLogger(__name__)


class NotImplementedManager(object):
    """Helper class to make any subclass of LoadBalancerBaseDriver explode if
    it is missing any of the required object managers.
    """

    def create(self, context, obj):
        raise NotImplementedError()

    def update(self, context, old_obj, obj):
        raise NotImplementedError()

    def delete(self, context, obj):
        raise NotImplementedError()


class EdgeLoadbalancerDriverV2(object):
    @log_helpers.log_method_call
    def __init__(self):
        super(EdgeLoadbalancerDriverV2, self).__init__()

        self.loadbalancer = lb_mgr.EdgeLoadBalancerManager()
        self.listener = listener_mgr.EdgeListenerManager()
        self.pool = pool_mgr.EdgePoolManager()
        self.member = member_mgr.EdgeMemberManager()
        self.healthmonitor = hm_mgr.EdgeHealthMonitorManager()
        self.l7policy = l7policy_mgr.EdgeL7PolicyManager()
        self.l7rule = l7rule_mgr.EdgeL7RuleManager()
        self._subscribe_router_delete_callback()

    def _subscribe_router_delete_callback(self):
        # Check if there is any LB attachment for the NSX router.
        # This callback is subscribed here to prevent router deletion
        # if it still has LB service attached to it.
        registry.subscribe(self._check_lb_service_on_router,
                           resources.ROUTER, events.BEFORE_DELETE)

    def _unsubscribe_router_delete_callback(self):
        registry.unsubscribe(self._check_lb_service_on_router,
                             resources.ROUTER, events.BEFORE_DELETE)

    def _check_lb_service_on_router(self, resource, event, trigger,
                                    **kwargs):
        """Check if there is any lb service on nsx router"""

        nsx_router_id = nsx_db.get_nsx_router_id(kwargs['context'].session,
                                                 kwargs['router_id'])
        nsxlib = self.loadbalancer.core_plugin.nsxlib
        service_client = nsxlib.load_balancer.service
        lb_service = service_client.get_router_lb_service(nsx_router_id)
        if lb_service:
            msg = _('Cannot delete router as it still has lb service '
                    'attachment')
            raise nc_exc.CallbackFailure(msg)


class DummyLoadbalancerDriverV2(object):
    @log_helpers.log_method_call
    def __init__(self):
        self.load_balancer = NotImplementedManager()
        self.listener = NotImplementedManager()
        self.pool = NotImplementedManager()
        self.member = NotImplementedManager()
        self.health_monitor = NotImplementedManager()
        self.l7policy = NotImplementedManager()
        self.l7rule = NotImplementedManager()



# Copyright 2012 VMware, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
from sqlalchemy.orm import exc

from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils

import neutron.db.api as db

from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import nsx_models

LOG = logging.getLogger(__name__)


def _apply_filters_to_query(query, model, filters, like_filters=None):
    if filters:
        for key, value in six.iteritems(filters):
            column = getattr(model, key, None)
            if column:
                query = query.filter(column.in_(value))
    if like_filters:
        for key, search_term in six.iteritems(like_filters):
            column = getattr(model, key, None)
            if column:
                query = query.filter(column.like(search_term))
    return query


def get_network_bindings(session, network_id):
    session = session or db.get_reader_session()
    return (session.query(nsx_models.TzNetworkBinding).
            filter_by(network_id=network_id).
            all())


def get_network_bindings_by_vlanid_and_physical_net(session, vlan_id,
                                                    phy_uuid):
    session = session or db.get_reader_session()
    return (session.query(nsx_models.TzNetworkBinding).
            filter_by(vlan_id=vlan_id, phy_uuid=phy_uuid).
            all())


def delete_network_bindings(session, network_id):
    return (session.query(nsx_models.TzNetworkBinding).
            filter_by(network_id=network_id).delete())


def add_network_binding(session, network_id, binding_type, phy_uuid, vlan_id):
    with session.begin(subtransactions=True):
        binding = nsx_models.TzNetworkBinding(network_id, binding_type,
                                              phy_uuid, vlan_id)
        session.add(binding)
    return binding


def add_neutron_nsx_network_mapping(session, neutron_id, nsx_switch_id,
                                    dvs_id=None):
    with session.begin(subtransactions=True):
        mapping = nsx_models.NeutronNsxNetworkMapping(
            neutron_id=neutron_id, nsx_id=nsx_switch_id,
            dvs_id=dvs_id)
        session.add(mapping)
        return mapping


def delete_neutron_nsx_network_mapping(session, neutron_id):
    return (session.query(nsx_models.NeutronNsxNetworkMapping).
            filter_by(neutron_id=neutron_id).delete())


def add_neutron_nsx_port_mapping(session, neutron_id,
                                 nsx_switch_id, nsx_port_id):
    session.begin(subtransactions=True)
    try:
        mapping = nsx_models.NeutronNsxPortMapping(
            neutron_id, nsx_switch_id, nsx_port_id)
        session.add(mapping)
        session.commit()
    except db_exc.DBDuplicateEntry:
        with excutils.save_and_reraise_exception() as ctxt:
            session.rollback()
            # do not complain if the same exact mapping is being added,
            # otherwise re-raise because even though it is possible for the
            # same neutron port to map to different back-end ports over time,
            # this should not occur whilst a mapping already exists
            current = get_nsx_switch_and_port_id(session, neutron_id)
            if current[1] == nsx_port_id:
                LOG.debug("Port mapping for %s already available",
                          neutron_id)
                ctxt.reraise = False
    except db_exc.DBError:
        with excutils.save_and_reraise_exception():
            # rollback for any other db error
            session.rollback()
    return mapping


def add_neutron_nsx_router_mapping(session, neutron_id, nsx_router_id):
    with session.begin(subtransactions=True):
        mapping = nsx_models.NeutronNsxRouterMapping(
            neutron_id=neutron_id, nsx_id=nsx_router_id)
        session.add(mapping)
        return mapping


def add_neutron_nsx_security_group_mapping(session, neutron_id, nsx_id):
    """Map a Neutron security group to a NSX security profile.

    :param session: a valid database session object
    :param neutron_id: a neutron security group identifier
    :param nsx_id: a nsx security profile identifier
    """
    with session.begin(subtransactions=True):
        mapping = nsx_models.NeutronNsxSecurityGroupMapping(
            neutron_id=neutron_id, nsx_id=nsx_id)
        session.add(mapping)
    return mapping


def get_nsx_service_binding(session, network_id, service_type):
    try:
        return session.query(nsx_models.NeutronNsxServiceBinding).filter_by(
            network_id=network_id, nsx_service_type=service_type).one()
    except exc.NoResultFound:
        LOG.debug("NSX %s service not enabled on network %s", service_type,
                  network_id)


def add_neutron_nsx_service_binding(session, network_id, port_id,
                                    service_type, service_id):
    """Store enabled NSX services on each Neutron network.

    :param session: database session object
    :param network_id: identifier of Neutron network enabling the service
    :param port_id: identifier of Neutron port providing the service
    :param service_type: type of NSX service
    :param service_id: identifier of NSX service
    """
    with session.begin(subtransactions=True):
        binding = nsx_models.NeutronNsxServiceBinding(
            network_id=network_id, port_id=port_id,
            nsx_service_type=service_type, nsx_service_id=service_id)
        session.add(binding)
        return binding


def delete_neutron_nsx_service_binding(session, network_id, service_type):
    return session.query(nsx_models.NeutronNsxServiceBinding).filter_by(
        network_id=network_id, nsx_service_type=service_type).delete()


def update_nsx_dhcp_bindings(session, port_id, org_ip, new_ip):
    try:
        with session.begin(subtransactions=True):
            binding = (session.query(nsx_models.NeutronNsxDhcpBinding).
                       filter_by(port_id=port_id, ip_address=org_ip).one())
            binding.ip_address = new_ip
    except exc.NoResultFound:
        LOG.debug("Binding not found for port %s", port_id)
        return


def get_nsx_dhcp_bindings(session, port_id):
    return [binding for binding in session.query(
        nsx_models.NeutronNsxDhcpBinding).filter_by(port_id=port_id)]


def get_nsx_dhcp_bindings_by_service(session, service_id):
    return [binding for binding in session.query(
        nsx_models.NeutronNsxDhcpBinding).filter_by(nsx_service_id=service_id)]


def add_neutron_nsx_dhcp_binding(session, port_id, subnet_id, ip_address,
                                 service_id, binding_id):
    """Store DHCP binding of each Neutron port.

    :param session: database session object
    :param port_id: identifier of Neutron port with DHCP binding
    :param subnet_id: identifier of Neutron subnet for the port
    :param ip_address: IP address for the port in this subnet.
    :param service_id: identifier of NSX DHCP service
    :param binding_id: identifier of NSX DHCP binding
    """
    with session.begin(subtransactions=True):
        binding = nsx_models.NeutronNsxDhcpBinding(
            port_id=port_id, subnet_id=subnet_id, ip_address=ip_address,
            nsx_service_id=service_id, nsx_binding_id=binding_id)
        session.add(binding)
        return binding


def delete_neutron_nsx_dhcp_binding(session, port_id, binding_id):
    return session.query(nsx_models.NeutronNsxDhcpBinding).filter_by(
        port_id=port_id, nsx_binding_id=binding_id).delete()


def delete_neutron_nsx_dhcp_bindings_by_service_id(session, service_id):
    return session.query(nsx_models.NeutronNsxDhcpBinding).filter_by(
        nsx_service_id=service_id).delete()


def get_nsx_switch_ids(session, neutron_id):
    # This function returns a list of NSX switch identifiers because of
    # the possibility of chained logical switches
    return [mapping['nsx_id'] for mapping in
            session.query(nsx_models.NeutronNsxNetworkMapping).filter_by(
                neutron_id=neutron_id)]


def get_nsx_network_mappings(session, neutron_id):
    # This function returns a list of NSX switch identifiers because of
    # the possibility of chained logical switches
    return session.query(nsx_models.NeutronNsxNetworkMapping).filter_by(
                neutron_id=neutron_id).all()


def get_nsx_switch_id_for_dvs(session, neutron_id, dvs_id):
    """Retrieve the NSX switch ID for a given DVS ID and neutron network."""
    try:
        mapping = (session.query(nsx_models.NeutronNsxNetworkMapping).
                   filter_by(neutron_id=neutron_id,
                             dvs_id=dvs_id).one())
        return mapping['nsx_id']
    except exc.NoResultFound:
        LOG.debug("NSX switch for dvs-id: %s not yet stored in Neutron DB",
                  dvs_id)


def get_net_ids(session, nsx_id):
    return [mapping['neutron_id'] for mapping in
            get_nsx_network_mapping_for_nsx_id(session, nsx_id)]


def get_nsx_network_mapping_for_nsx_id(session, nsx_id):
    return session.query(nsx_models.NeutronNsxNetworkMapping).filter_by(
        nsx_id=nsx_id).all()


def get_nsx_networks_mapping(session):
    return session.query(nsx_models.NeutronNsxNetworkMapping).all()


def get_nsx_switch_and_port_id(session, neutron_id):
    try:
        mapping = (session.query(nsx_models.NeutronNsxPortMapping).
                   filter_by(neutron_id=neutron_id).
                   one())
        return mapping['nsx_switch_id'], mapping['nsx_port_id']
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron port %s not yet "
                  "stored in Neutron DB", neutron_id)
        return None, None


def get_nsx_router_id(session, neutron_id):
    try:
        mapping = (session.query(nsx_models.NeutronNsxRouterMapping).
                   filter_by(neutron_id=neutron_id).one())
        return mapping['nsx_id']
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron router %s not yet "
                  "stored in Neutron DB", neutron_id)


def get_neutron_from_nsx_router_id(session, nsx_router_id):
    try:
        mapping = (session.query(nsx_models.NeutronNsxRouterMapping).
                   filter_by(nsx_id=nsx_router_id).one())
        return mapping['neutron_id']
    except exc.NoResultFound:
        LOG.debug("Couldn't find router with nsx id  %s", nsx_router_id)


def get_nsx_security_group_id(session, neutron_id, moref=False):
    """Return the id of a security group in the NSX backend.

    Note: security groups are called 'security profiles' in NSX
    """
    try:
        mappings = (session.query(nsx_models.NeutronNsxSecurityGroupMapping).
                    filter_by(neutron_id=neutron_id).
                    all())
        for mapping in mappings:
            if moref and not uuidutils.is_uuid_like(mapping['nsx_id']):
                return mapping['nsx_id']
            if not moref and uuidutils.is_uuid_like(mapping['nsx_id']):
                return mapping['nsx_id']
    except exc.NoResultFound:
        LOG.debug("NSX identifiers for neutron security group %s not yet "
                  "stored in Neutron DB", neutron_id)
        return None


def get_nsx_security_group_ids(session, neutron_ids):
    """Return list of ids of a security groups in the NSX backend.
    """
    filters = {'neutron_id': neutron_ids}
    like_filters = None
    query = session.query(nsx_models.NeutronNsxSecurityGroupMapping)
    mappings = _apply_filters_to_query(
        query, nsx_models.NeutronNsxSecurityGroupMapping,
        filters, like_filters).all()
    return [mapping['nsx_id'] for mapping in mappings
            if mapping['nsx_id'] is not None]


def _delete_by_neutron_id(session, model, neutron_id):
    return session.query(model).filter_by(neutron_id=neutron_id).delete()


def delete_neutron_nsx_port_mapping(session, neutron_id):
    return _delete_by_neutron_id(
        session, nsx_models.NeutronNsxPortMapping, neutron_id)


def delete_neutron_nsx_router_mapping(session, neutron_id):
    return _delete_by_neutron_id(
        session, nsx_models.NeutronNsxRouterMapping, neutron_id)


def unset_default_network_gateways(session):
    with session.begin(subtransactions=True):
        session.query(nsx_models.NetworkGateway).update(
            {nsx_models.NetworkGateway.default: False})


def set_default_network_gateway(session, gw_id):
    with session.begin(subtransactions=True):
        gw = (session.query(nsx_models.NetworkGateway).
              filter_by(id=gw_id).one())
        gw['default'] = True


def set_multiprovider_network(session, network_id):
    with session.begin(subtransactions=True):
        multiprovider_network = nsx_models.MultiProviderNetworks(
            network_id)
        session.add(multiprovider_network)
        return multiprovider_network


def is_multiprovider_network(session, network_id):
    with session.begin(subtransactions=True):
        return bool(
            session.query(nsx_models.MultiProviderNetworks).filter_by(
                network_id=network_id).first())


# NSXv3 L2 Gateway DB methods.
def add_l2gw_connection_mapping(session, connection_id, bridge_endpoint_id,
                                port_id):
    with session.begin(subtransactions=True):
        mapping = nsx_models.NsxL2GWConnectionMapping(
            connection_id=connection_id,
            port_id=port_id,
            bridge_endpoint_id=bridge_endpoint_id)
        session.add(mapping)
        return mapping


def get_l2gw_connection_mapping(session, connection_id):
    try:
        return (session.query(nsx_models.NsxL2GWConnectionMapping).
                filter_by(connection_id=connection_id).one())
    except exc.NoResultFound:
        raise nsx_exc.NsxL2GWConnectionMappingNotFound(conn=connection_id)


# NSXv3 QoS policy id <-> switch Id mapping
def add_qos_policy_profile_mapping(session, qos_policy_id, switch_profile_id):
    with session.begin(subtransactions=True):
        mapping = nsx_models.QosPolicySwitchProfile(
            qos_policy_id=qos_policy_id,
            switch_profile_id=switch_profile_id)
        session.add(mapping)
        return mapping


def get_switch_profile_by_qos_policy(session, qos_policy_id):
    try:
        entry = (session.query(nsx_models.QosPolicySwitchProfile).
                 filter_by(qos_policy_id=qos_policy_id).one())
        return entry.switch_profile_id
    except exc.NoResultFound:
        raise nsx_exc.NsxQosPolicyMappingNotFound(policy=qos_policy_id)


def delete_qos_policy_profile_mapping(session, qos_policy_id):
    return (session.query(nsx_models.QosPolicySwitchProfile).
            filter_by(qos_policy_id=qos_policy_id).delete())


# NSXv3 Port Mirror Sessions DB methods.
def add_port_mirror_session_mapping(session, tf_id, pm_session_id):
    with session.begin(subtransactions=True):
        mapping = nsx_models.NsxPortMirrorSessionMapping(
            tap_flow_id=tf_id,
            port_mirror_session_id=pm_session_id)
        session.add(mapping)
        return mapping


def get_port_mirror_session_mapping(session, tf_id):
    try:
        return (session.query(nsx_models.NsxPortMirrorSessionMapping).
                filter_by(tap_flow_id=tf_id).one())
    except exc.NoResultFound:
        raise nsx_exc.NsxPortMirrorSessionMappingNotFound(tf=tf_id)


def delete_port_mirror_session_mapping(session, tf_id):
    return (session.query(nsx_models.NsxPortMirrorSessionMapping).
            filter_by(tap_flow_id=tf_id).delete())


@db.context_manager.writer
def save_sg_mappings(context, sg_id, nsgroup_id, section_id):
    context.session.add(
        nsx_models.NeutronNsxFirewallSectionMapping(neutron_id=sg_id,
                                                    nsx_id=section_id))
    context.session.add(
        nsx_models.NeutronNsxSecurityGroupMapping(neutron_id=sg_id,
                                                  nsx_id=nsgroup_id))


def get_sg_mappings(session, sg_id, moref=False):
    nsgroup_mappings = session.query(
        nsx_models.NeutronNsxSecurityGroupMapping
    ).filter_by(neutron_id=sg_id).all()
    nsgroup_mapping = section_mapping = None
    for mapping in nsgroup_mappings:
        if moref and not uuidutils.is_uuid_like(mapping['nsx_id']):
            nsgroup_mapping = mapping['nsx_id']
            break
        if not moref and uuidutils.is_uuid_like(mapping['nsx_id']):
            nsgroup_mapping = mapping['nsx_id']
            break
    section_mappings = session.query(
        nsx_models.NeutronNsxFirewallSectionMapping
    ).filter_by(neutron_id=sg_id).all()
    for mapping in section_mappings:
        if moref and not uuidutils.is_uuid_like(mapping['nsx_id']):
            section_mapping = mapping['nsx_id']
            break
        if not moref and uuidutils.is_uuid_like(mapping['nsx_id']):
            section_mapping = mapping['nsx_id']
            break
    return nsgroup_mapping, section_mapping


def get_sg_rule_mapping(session, rule_id):
    rule_mapping = session.query(
        nsx_models.NeutronNsxRuleMapping).filter_by(
        neutron_id=rule_id).one()
    return rule_mapping.nsx_id


def save_sg_rule_mappings(session, rules):
    with session.begin(subtransactions=True):
        for neutron_id, nsx_id in rules:
            mapping = nsx_models.NeutronNsxRuleMapping(
                neutron_id=neutron_id, nsx_id=nsx_id)
            session.add(mapping)


def add_nsx_ipam_subnet_pool(session, subnet_id, nsx_pool_id):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxSubnetIpam(
            subnet_id=subnet_id,
            nsx_pool_id=nsx_pool_id)
        session.add(binding)
    return binding


def get_nsx_ipam_pool_for_subnet(session, subnet_id):
    try:
        entry = session.query(
            nsx_models.NsxSubnetIpam).filter_by(
            subnet_id=subnet_id).one()
        return entry.nsx_pool_id
    except exc.NoResultFound:
        return


def del_nsx_ipam_subnet_pool(session, subnet_id, nsx_pool_id):
    return (session.query(nsx_models.NsxSubnetIpam).
            filter_by(subnet_id=subnet_id,
                      nsx_pool_id=nsx_pool_id).delete())


def get_certificate(session, purpose):
    try:
        cert_entry = session.query(
            nsx_models.NsxCertificateRepository).filter_by(
                purpose=purpose).one()
        return cert_entry.certificate, cert_entry.private_key
    except exc.NoResultFound:
        return None, None


def save_certificate(session, purpose, cert, pk):
    with session.begin(subtransactions=True):
        cert_entry = nsx_models.NsxCertificateRepository(
                purpose=purpose,
                certificate=cert,
                private_key=pk)
        session.add(cert_entry)


def delete_certificate(session, purpose):
    return (session.query(nsx_models.NsxCertificateRepository).
            filter_by(purpose=purpose).delete())


def add_nsx_lbaas_loadbalancer_binding(session, loadbalancer_id,
                                       lb_service_id, lb_router_id,
                                       vip_address):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxLbaasLoadbalancer(
            loadbalancer_id=loadbalancer_id, lb_service_id=lb_service_id,
            lb_router_id=lb_router_id, vip_address=vip_address)
        session.add(binding)
    return binding


def get_nsx_lbaas_loadbalancer_binding(session, loadbalancer_id):
    try:
        return session.query(
            nsx_models.NsxLbaasLoadbalancer).filter_by(
                loadbalancer_id=loadbalancer_id).one()
    except exc.NoResultFound:
        return


def get_nsx_lbaas_loadbalancer_binding_by_service(session, lb_service_id):
    return session.query(
        nsx_models.NsxLbaasLoadbalancer).filter_by(
            lb_service_id=lb_service_id).all()


def delete_nsx_lbaas_loadbalancer_binding(session, loadbalancer_id):
    return (session.query(nsx_models.NsxLbaasLoadbalancer).
            filter_by(loadbalancer_id=loadbalancer_id).delete())


def add_nsx_lbaas_listener_binding(session, loadbalancer_id, listener_id,
                                   app_profile_id, lb_vs_id):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxLbaasListener(
            loadbalancer_id=loadbalancer_id, listener_id=listener_id,
            app_profile_id=app_profile_id,
            lb_vs_id=lb_vs_id)
        session.add(binding)
    return binding


def get_nsx_lbaas_listener_binding(session, loadbalancer_id, listener_id):
    try:
        return session.query(
            nsx_models.NsxLbaasListener).filter_by(
                loadbalancer_id=loadbalancer_id,
                listener_id=listener_id).one()
    except exc.NoResultFound:
        return


def delete_nsx_lbaas_listener_binding(session, loadbalancer_id, listener_id):
    return (session.query(nsx_models.NsxLbaasListener).
            filter_by(loadbalancer_id=loadbalancer_id,
                      listener_id=listener_id).delete())


def add_nsx_lbaas_pool_binding(session, loadbalancer_id, pool_id, lb_pool_id,
                               lb_vs_id=None):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxLbaasPool(loadbalancer_id=loadbalancer_id,
                                          pool_id=pool_id,
                                          lb_pool_id=lb_pool_id,
                                          lb_vs_id=lb_vs_id)
        session.add(binding)
    return binding


def get_nsx_lbaas_pool_binding(session, loadbalancer_id, pool_id):
    try:
        return session.query(nsx_models.NsxLbaasPool).filter_by(
            loadbalancer_id=loadbalancer_id, pool_id=pool_id).one()
    except exc.NoResultFound:
        return


def update_nsx_lbaas_pool_binding(session, loadbalancer_id, pool_id,
                                  lb_vs_id):
    try:
        with session.begin(subtransactions=True):
            binding = (session.query(nsx_models.NsxLbaasPool).
                       filter_by(loadbalancer_id=loadbalancer_id,
                                 pool_id=pool_id).one())
            binding.lb_vs_id = lb_vs_id
    except exc.NoResultFound:
        LOG.debug("Binding not found for pool %s", pool_id)
        return


def delete_nsx_lbaas_pool_binding(session, loadbalancer_id, pool_id):
    return (session.query(nsx_models.NsxLbaasPool).
            filter_by(loadbalancer_id=loadbalancer_id,
                      pool_id=pool_id).delete())


def add_nsx_lbaas_monitor_binding(session, loadbalancer_id, pool_id, hm_id,
                                  lb_monitor_id, lb_pool_id):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxLbaasMonitor(
            loadbalancer_id=loadbalancer_id, pool_id=pool_id, hm_id=hm_id,
            lb_monitor_id=lb_monitor_id, lb_pool_id=lb_pool_id)
        session.add(binding)
    return binding


def get_nsx_lbaas_monitor_binding(session, loadbalancer_id, pool_id, hm_id):
    try:
        return session.query(nsx_models.NsxLbaasMonitor).filter_by(
            loadbalancer_id=loadbalancer_id,
            pool_id=pool_id, hm_id=hm_id).one()
    except exc.NoResultFound:
        return


def delete_nsx_lbaas_monitor_binding(session, loadbalancer_id, pool_id,
                                     hm_id):
    return (session.query(nsx_models.NsxLbaasMonitor).
            filter_by(loadbalancer_id=loadbalancer_id,
                      pool_id=pool_id, hm_id=hm_id).delete())


def add_nsx_lbaas_l7policy_binding(session, l7policy_id, lb_rule_id, lb_vs_id):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxLbaasL7Policy(
            l7policy_id=l7policy_id, lb_rule_id=lb_rule_id, lb_vs_id=lb_vs_id)
        session.add(binding)
    return binding


def get_nsx_lbaas_l7policy_binding(session, l7policy_id):
    try:
        return session.query(nsx_models.NsxLbaasL7Policy).filter_by(
            l7policy_id=l7policy_id).one()
    except exc.NoResultFound:
        return


def delete_nsx_lbaas_l7policy_binding(session, l7policy_id):
    return (session.query(nsx_models.NsxLbaasL7Policy).
            filter_by(l7policy_id=l7policy_id).delete())


def add_project_plugin_mapping(session, project, plugin):
    with session.begin(subtransactions=True):
        binding = nsx_models.NsxProjectPluginMapping(
            project=project, plugin=plugin)
        session.add(binding)
    return binding


def get_project_plugin_mapping(session, project):
    try:
        return session.query(nsx_models.NsxProjectPluginMapping).filter_by(
            project=project).one()
    except exc.NoResultFound:
        return


def get_project_plugin_mappings(session):
    return session.query(nsx_models.NsxProjectPluginMapping).all()


def get_project_plugin_mappings_by_plugin(session, plugin):
    return session.query(nsx_models.NsxProjectPluginMapping).filter_by(
        plugin=plugin).all()


def add_nsx_vpn_connection_mapping(session, neutron_id, session_id,
                                   dpd_profile_id, ike_profile_id,
                                   ipsec_profile_id, peer_ep_id):
    with session.begin(subtransactions=True):
        mapping = nsx_models.NsxVpnConnectionMapping(
            neutron_id=neutron_id,
            session_id=session_id,
            dpd_profile_id=dpd_profile_id,
            ike_profile_id=ike_profile_id,
            ipsec_profile_id=ipsec_profile_id,
            peer_ep_id=peer_ep_id)
        session.add(mapping)
        return mapping


def get_nsx_vpn_connection_mapping(session, neutron_id):
    try:
        return (session.query(nsx_models.NsxVpnConnectionMapping).
            filter_by(neutron_id=neutron_id).one())
    except exc.NoResultFound:
        return


def delete_nsx_vpn_connection_mapping(session, neutron_id):
    return (session.query(nsx_models.NsxVpnConnectionMapping).
            filter_by(neutron_id=neutron_id).delete())



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import xml.etree.ElementTree as et

from oslo_log import log as logging

from vmware_nsx.common import utils

WAIT_INTERVAL = 2000
MAX_ATTEMPTS = 5

LOG = logging.getLogger(__name__)


class NsxSecurityGroupUtils(object):

    def __init__(self, nsxv_manager):
        LOG.debug("Start Security Group Utils initialization")
        self.nsxv_manager = nsxv_manager

    def to_xml_string(self, element):
        return et.tostring(element)

    def get_section_with_rules(self, name, rules, section_id=None):
        """Helper method to create section dict with rules."""

        section = et.Element('section')
        section.attrib['name'] = name
        if section_id:
            section.attrib['id'] = section_id
        for rule in rules:
            section.append(rule)
        return section

    def get_container(self, nsx_sg_id):
        container = {'type': 'SecurityGroup', 'value': nsx_sg_id}
        return container

    def get_remote_container(self, remote_group_id, remote_ip_mac):
        container = None
        if remote_group_id is not None:
            return self.get_container(remote_group_id)
        if remote_ip_mac is not None:
            container = {'type': 'Ipv4Address', 'value': remote_ip_mac}
        return container

    def get_rule_config(self, applied_to_ids, name, action='allow',
                        applied_to='SecurityGroup',
                        source=None, destination=None, services=None,
                        flags=None, logged=False, tag=None,
                        application_services=None):
        """Helper method to create a nsx rule dict."""
        ruleTag = et.Element('rule')
        ruleTag.attrib['logged'] = 'true' if logged else 'false'
        nameTag = et.SubElement(ruleTag, 'name')
        nameTag.text = name
        actionTag = et.SubElement(ruleTag, 'action')
        actionTag.text = action

        apList = et.SubElement(ruleTag, 'appliedToList')
        for applied_to_id in applied_to_ids:
            apTag = et.SubElement(apList, 'appliedTo')
            apTypeTag = et.SubElement(apTag, 'type')
            apTypeTag.text = applied_to
            apValueTag = et.SubElement(apTag, 'value')
            apValueTag.text = applied_to_id

        if source is not None:
            sources = et.SubElement(ruleTag, 'sources')
            sources.attrib['excluded'] = 'false'
            srcTag = et.SubElement(sources, 'source')
            srcTypeTag = et.SubElement(srcTag, 'type')
            srcTypeTag.text = source['type']
            srcValueTag = et.SubElement(srcTag, 'value')
            srcValueTag.text = source['value']

        if destination is not None:
            dests = et.SubElement(ruleTag, 'destinations')
            dests.attrib['excluded'] = 'false'
            destTag = et.SubElement(dests, 'destination')
            destTypeTag = et.SubElement(destTag, 'type')
            destTypeTag.text = destination['type']
            destValueTag = et.SubElement(destTag, 'value')
            destValueTag.text = destination['value']

        if services:
            s = et.SubElement(ruleTag, 'services')
            for protocol, port, icmptype, icmpcode in services:
                svcTag = et.SubElement(s, 'service')
                try:
                    int(protocol)
                    svcProtocolTag = et.SubElement(svcTag, 'protocol')
                    svcProtocolTag.text = str(protocol)
                except ValueError:
                    svcProtocolTag = et.SubElement(svcTag, 'protocolName')
                    svcProtocolTag.text = protocol
                if port is not None:
                    svcPortTag = et.SubElement(svcTag, 'destinationPort')
                    svcPortTag.text = str(port)
                if icmptype is not None:
                    svcPortTag = et.SubElement(svcTag, 'subProtocol')
                    svcPortTag.text = str(icmptype)
                if icmpcode is not None:
                    svcPortTag = et.SubElement(svcTag, 'icmpCode')
                    svcPortTag.text = str(icmpcode)

        if application_services:
            s = et.SubElement(ruleTag, 'services')
            for application_service in application_services:
                svcTag = et.SubElement(s, 'service')
                svcProtocolTag = et.SubElement(svcTag, 'value')
                svcProtocolTag.text = str(application_service)

        if flags:
            if flags.get('ethertype') is not None:
                pktTag = et.SubElement(ruleTag, 'packetType')
                pktTag.text = flags.get('ethertype')
            if flags.get('direction') is not None:
                dirTag = et.SubElement(ruleTag, 'direction')
                dirTag.text = flags.get('direction')
        if tag:
            tagTag = et.SubElement(ruleTag, 'tag')
            tagTag.text = tag
        return ruleTag

    def get_rule_id_pair_from_section(self, resp):
        root = et.fromstring(resp)
        pairs = []
        for rule in root.findall('rule'):
            pair = {'nsx_id': rule.attrib.get('id'),
                    'neutron_id': rule.find('name').text}
            pairs.append(pair)
        return pairs

    def extend_section_with_rules(self, section, nsx_rules):
        section.extend(nsx_rules)

    def parse_section(self, xml_string):
        return et.fromstring(xml_string)

    def get_nsx_sg_name(self, sg_data):
        return '%(name)s (%(id)s)' % sg_data

    def get_nsx_section_name(self, sg_data):
        return 'SG Section: %s' % self.get_nsx_sg_name(sg_data)

    def parse_and_get_section_id(self, section_xml):
        section = et.fromstring(section_xml)
        return section.attrib['id']

    def is_section_logged(self, section):
        # Determine if this section rules are being logged by the first rule
        # 'logged' value.
        rule = section.find('rule')
        if rule is not None:
            return rule.attrib.get('logged') == 'true'
        return False

    def set_rules_logged_option(self, section, logged):
        value = 'true' if logged else 'false'
        rules = section.findall('rule')
        updated = False
        for rule in rules:
            if rule.attrib['logged'] != value:
                rule.attrib['logged'] = value
                updated = True
        return updated

    def del_nsx_security_group_from_policy(self, policy_id, sg_id):
        if not policy_id:
            return
        policy = self.nsxv_manager.vcns.get_security_policy(policy_id)
        policy = utils.normalize_xml(policy)

        # check if the security group is already bounded to the policy
        for binding in policy.iter('securityGroupBinding'):
            if binding.find('objectId').text == sg_id:
                # delete this entry
                policy.remove(binding)

                return self.nsxv_manager.vcns.update_security_policy(
                    policy_id, et.tostring(policy))

    def add_nsx_security_group_to_policy(self, policy_id, sg_id):
        if not policy_id:
            return
        # Get the policy configuration
        policy = self.nsxv_manager.vcns.get_security_policy(policy_id)
        policy = utils.normalize_xml(policy)

        # check if the security group is already bounded to the policy
        for binding in policy.iter('securityGroupBinding'):
            if binding.find('objectId').text == sg_id:
                # Already there
                return

        # Add a new binding entry
        new_binding = et.SubElement(policy, 'securityGroupBinding')
        et.SubElement(new_binding, 'objectId').text = sg_id

        return self.nsxv_manager.vcns.update_security_policy(
            policy_id, et.tostring(policy))



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from neutron_lib.api import validators
from neutron_lib import constants as n_consts
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from neutron_lib.services.qos import constants as qos_consts

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import db as nsx_db
from vmware_nsx.extensions import projectpluginmap

LOG = logging.getLogger(__name__)

MAX_KBPS_MIN_VALUE = 1024
# The max limit is calculated so that the value sent to the backed will
# be smaller than 2**31
MAX_BURST_MAX_VALUE = int((2 ** 31 - 1) / 128)


class QosNotificationsHandler(object):

    def __init__(self):
        super(QosNotificationsHandler, self).__init__()
        self._core_plugin = None

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
            if self._core_plugin.is_tvd_plugin():
                # get the plugin that match this driver
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_T)
        return self._core_plugin

    @property
    def _nsxlib_qos(self):
        return self.core_plugin.nsxlib.qos_switching_profile

    def _get_tags(self, context, policy):
        policy_dict = {'id': policy.id, 'tenant_id': policy.tenant_id}
        return self._nsxlib_qos.build_v3_tags_payload(
            policy_dict, resource_type='os-neutron-qos-id',
            project_name=context.tenant_name)

    def create_policy(self, context, policy):
        policy_id = policy.id
        tags = self._get_tags(context, policy)
        result = self._nsxlib_qos.create(
            tags=tags, name=policy.name,
            description=policy.description)
        if not result or not validators.is_attr_set(result.get('id')):
            msg = _("Unable to create QoS switching profile on the backend")
            raise nsx_exc.NsxPluginException(err_msg=msg)
        profile_id = result['id']

        # Add the mapping entry of the policy_id <-> profile_id
        nsx_db.add_qos_policy_profile_mapping(context.session,
                                              policy_id,
                                              profile_id)

    def delete_policy(self, context, policy_id):
        profile_id = nsx_db.get_switch_profile_by_qos_policy(
            context.session, policy_id)
        # delete the profile id from the backend and the DB
        self._nsxlib_qos.delete(profile_id)
        nsx_db.delete_qos_policy_profile_mapping(
            context.session, policy_id)

    def update_policy(self, context, policy_id, policy):
        profile_id = nsx_db.get_switch_profile_by_qos_policy(
            context.session, policy_id)
        tags = self._get_tags(context, policy)
        self._nsxlib_qos.update(
            profile_id,
            tags=tags,
            name=policy.name,
            description=policy.description)

    def _validate_bw_values(self, bw_rule):
        """Validate that the configured values are allowed by the NSX backend.

        Since failing the action from the notification callback
        is not possible, just log the warning and use the minimal/maximal
        values.
        """
        # Validate the max bandwidth value minimum value
        # (max value is above what neutron allows so no need to check it)
        if (bw_rule.max_kbps < MAX_KBPS_MIN_VALUE):
            msg = (_("Invalid input for max_kbps. "
                     "The minimal legal value is %s") % MAX_KBPS_MIN_VALUE)
            LOG.error(msg)
            raise n_exc.InvalidInput(error_message=msg)

        # validate the burst size value max value
        # (max value is 0, and neutron already validates this)
        if (bw_rule.max_burst_kbps > MAX_BURST_MAX_VALUE):
            msg = (_("Invalid input for burst_size. "
                     "The maximal legal value is %s") % MAX_BURST_MAX_VALUE)
            LOG.error(msg)
            raise n_exc.InvalidInput(error_message=msg)

    def _get_bw_values_from_rule(self, bw_rule):
        """Translate the neutron bandwidth_limit_rule values, into the
        values expected by the NSX-v3 QoS switch profile,
        and validate that those are legal
        """
        if bw_rule:
            shaping_enabled = True

            # translate kbps -> bytes
            burst_size = int(bw_rule.max_burst_kbps) * 128

            # translate kbps -> Mbps
            average_bandwidth = int(round(float(bw_rule.max_kbps) / 1024))

            # peakBandwidth: a Multiplying on the average BW
            # because the neutron qos configuration supports
            # only 1 value
            peak_bandwidth = int(round(average_bandwidth *
                                       cfg.CONF.NSX.qos_peak_bw_multiplier))
        else:
            shaping_enabled = False
            burst_size = None
            peak_bandwidth = None
            average_bandwidth = None

        return shaping_enabled, burst_size, peak_bandwidth, average_bandwidth

    def _get_dscp_values_from_rule(self, dscp_rule):
        """Translate the neutron DSCP marking rule values, into the
        values expected by the NSX-v3 QoS switch profile
        """
        if dscp_rule:
            qos_marking = 'untrusted'
            dscp = dscp_rule.dscp_mark
        else:
            qos_marking = 'trusted'
            dscp = 0

        return qos_marking, dscp

    def update_policy_rules(self, context, policy_id, rules):
        """Update the QoS switch profile with the BW limitations and
        DSCP marking configuration
        """
        profile_id = nsx_db.get_switch_profile_by_qos_policy(
            context.session, policy_id)

        ingress_bw_rule = None
        egress_bw_rule = None
        dscp_rule = None
        for rule in rules:
            if rule.rule_type == qos_consts.RULE_TYPE_BANDWIDTH_LIMIT:
                if rule.direction == n_consts.EGRESS_DIRECTION:
                    egress_bw_rule = rule
                else:
                    ingress_bw_rule = rule
            else:
                dscp_rule = rule

        # the NSX direction is opposite to the neutron direction
        (ingress_bw_enabled, ingress_burst_size, ingress_peak_bw,
            ingress_average_bw) = self._get_bw_values_from_rule(egress_bw_rule)

        (egress_bw_enabled, egress_burst_size, egress_peak_bw,
            egress_average_bw) = self._get_bw_values_from_rule(ingress_bw_rule)

        qos_marking, dscp = self._get_dscp_values_from_rule(dscp_rule)

        self._nsxlib_qos.set_profile_shaping(
            profile_id,
            ingress_bw_enabled=ingress_bw_enabled,
            ingress_burst_size=ingress_burst_size,
            ingress_peak_bandwidth=ingress_peak_bw,
            ingress_average_bandwidth=ingress_average_bw,
            egress_bw_enabled=egress_bw_enabled,
            egress_burst_size=egress_burst_size,
            egress_peak_bandwidth=egress_peak_bw,
            egress_average_bandwidth=egress_average_bw,
            qos_marking=qos_marking, dscp=dscp)

    def validate_policy_rule(self, context, policy_id, rule):
        """Raise an exception if the rule values are not supported"""
        if rule.rule_type == qos_consts.RULE_TYPE_BANDWIDTH_LIMIT:
            self._validate_bw_values(rule)



# Copyright 2015 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import six
from vmware_nsx._i18n import _
from vmware_nsx.db import db
from vmware_nsx.shell import resources as nsxadmin

from neutron.common import profiler  # noqa
from neutron_lib.callbacks import registry
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def output_header(func):
    """Decorator to demarcate the output of various hooks.

    Based on the callback function name we add a header to the
    cli output. Callback name's should follow the convention of
    component_operation_it_does to leverage the decorator
    """
    def func_desc(*args, **kwargs):
        component = '[%s]' % func.__name__.split('_')[0].upper()
        op_desc = [n.capitalize() for n in func.__name__.split('_')[1:]]
        LOG.info('==== %(component)s %(operation)s ====',
                 {'component': component, 'operation': ' '.join(op_desc)})
        return func(*args, **kwargs)
    func_desc.__name__ = func.__name__
    return func_desc


def parse_multi_keyval_opt(opt_list):
    """Converts a MutliStrOpt to a key-value dict"""

    result = dict()
    opt_list = opt_list if opt_list else []
    for opt_value in opt_list:
        try:
            key, value = opt_value.split('=')
            result[key] = value
        except ValueError:
            raise ValueError(_("Illegal argument [%s]: input should have the "
                               "format of '--property key=value'") % opt_value)
    return result


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError(_("invalid default answer: '%s'") % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = six.moves.input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def list_handler(resource):
    def wrap(func):
        registry.subscribe(func, resource,
                           nsxadmin.Operations.LIST.value)
        return func
    return wrap


def list_mismatches_handler(resource):
    def wrap(func):
        registry.subscribe(func, resource,
                           nsxadmin.Operations.LIST_MISMATCHES.value)
        return func
    return wrap


def fix_mismatches_handler(resource):
    def wrap(func):
        registry.subscribe(func, resource,
                           nsxadmin.Operations.FIX_MISMATCH.value)
        return func
    return wrap


def get_plugin_filters(context, plugin):
    # Return filters for the neutron list apis so that only resources from
    # a specific plugin will be returned.
    filters = {}
    core_plugin = nsxadmin.get_plugin()
    if core_plugin == 'nsxtvd':
        maps = db.get_project_plugin_mappings_by_plugin(
            context.session, plugin)
        if maps:
            filters['project_id'] = [m.project for m in maps]
    return filters
